"""Mixin classes for Renderers that plot.
"""

import os
import sys
import re
import math
import tempfile
import itertools
import numpy

from CGATReport.ResultBlock import ResultBlock, ResultBlocks
from CGATReportPlugins.Renderer import Renderer, NumpyMatrix
from CGATReport.DataTree import path2str
from CGATReport import Stats, Utils

try:
    from rpy2.robjects import r as R
    import rpy2.robjects as ro
    import rpy2.robjects.pandas2ri
    rpy2.robjects.pandas2ri.activate()
    import rpy2.robjects.numpy2ri
    rpy2.robjects.numpy2ri.activate()
except ImportError:
    R = None

from docutils.parsers.rst import directives


def parseRanges(r):
    '''given a string in the format "x,y",
    return a tuple of values (x,y).

    missing values are set to None.
    '''

    if not r:
        return r
    r = [x.strip() for x in r.split(",")]
    if r[0] == "":
        r[0] = None
    else:
        r[0] = float(r[0])
    if r[1] == "":
        r[1] = None
    else:
        r[1] = float(r[1])
    return r


def getCurrentRDevice():
    '''return the numerical device id of the current device.'''
    # R.dev_off()
    # return R.dev_cur().values()[0]
    return R["dev.cur"]()[0]


class Plotter(object):

    """Base class for Renderers that do simple 2D plotting.

    This mixin class provides convenience function
    for:class:`Renderer.Renderer` classes that do 2D plotting.

    The base class takes care of counting the plots created,
    provided that the methods:meth:`startPlot` and:meth:`endPlot`
    are called appropriately. It then inserts the appropriate place holders.

    This class adds the following options to the:term:`report` directive:

:term:`logscale`: apply logscales one or more axes.

:term:`xtitle`: add a label to the X axis

:term:`ytitle`: add a label to the Y axis

:term:`title`:  title of the plot

:term:`add-title`: add title to each plot

:term:`legend-location`: specify the location of the legend

:term:`as-lines`: do not plot symbols

:term:`xrange`: restrict plot a part of the x-axis

:term:`yrange`: restrict plot a part of the y-axis

    With some plots default layout options will result in plots
    that are misaligned (legends truncated, etc.). To fix this it might
    be necessary to increase plot size, reduce font size, or others.
    The following options will be passed on the matplotlib to permit
    this control.

:term:`mpl-figure`: options for matplotlib
           ``figure`` calls().

:term:`mpl-legend`: options for matplotlib
           ``legend`` calls().

:term:`mpl-subplot`: options for matplotlib
           ``subplots_adjust`` calls().

:term:`mpl-rc`: general environment settings for matplotlib.
          See the matplotlib documentation. Multiple options can be
          separated by ;, for example
          ``:mpl-rc: figure.figsize=(20,10);legend.fontsize=4``

    """

    mLegendFontSize = 8
    # number of chars to use to reduce legend font size
    mMaxLegendSize = 100

    # maximum number of rows per column. If there are more,
    # the legend is split into multiple columns
    mLegendMaxRowsPerColumn = 30

    options = (
        ('logscale',  directives.unchanged),
        ('title',  directives.unchanged),
        ('add-title',  directives.flag),
        ('xtitle',  directives.unchanged),
        ('ytitle',  directives.unchanged),
        ('xrange',  directives.unchanged),
        ('yrange',  directives.unchanged),
        ('zrange', directives.unchanged),
        ('mpl-figure',  directives.unchanged),
        ('mpl-legend',  directives.unchanged),
        ('mpl-subplot',  directives.unchanged),
        ('mpl-rc',  directives.unchanged),
        ('as-lines', directives.flag),
        ('legend-location',  directives.unchanged),
    )

    def __init__(self, *args, **kwargs):
        """parse option arguments."""

        self.mFigure = 0
        self.mColors = "bgrcmk"
        self.mSymbols = ["g-D", "b-h", "r-+", "c-+", "m-+",
                         "y-+", "k-o", "g-^", "b-<", "r->", "c-D", "m-h"]
        self.mMarkers = "so^>dph8+x"

        self.logscale = kwargs.get("logscale", None)
        self.title = kwargs.get("title", None)
        self.add_title = "add-title" in kwargs
        self.xlabel = kwargs.get("xtitle", None)
        self.ylabel = kwargs.get("ytitle", None)

        # substitute '-' in CGATReport-speak for ' ' in matplotlib speak
        self.legend_location = re.sub(
            "-", " ", kwargs.get("legend-location", "outer-top"))

        self.width = kwargs.get("width", 0.50)
        self.mAsLines = "as-lines" in kwargs
        self.xrange = parseRanges(kwargs.get("xrange", None))
        self.yrange = parseRanges(kwargs.get("yrange", None))
        self.zrange = parseRanges(kwargs.get("zrange", None))

        if self.mAsLines:
            self.mSymbols = []
            for y in ("-", ":", "--"):
                for x in "gbrcmyk":
                    self.mSymbols.append(y + x)

    def startPlot(self):
        R.x11()

    def endPlot(self, work, path):
        # currently: collects only single plots.
        figid = getCurrentRDevice()
        blocks = ResultBlocks(
            ResultBlock("\n".join(("#$rpl %i$#" % (figid), "")),
                        title=path2str(path)))
        return blocks


class LinePlot(Renderer, Plotter):

    '''create a line plot.

    This:class:`Renderer` requires at least three levels:

    line / data / coords.

    This is a base class that provides several hooks for
    derived classes.

    initPlot()

    for line, data in work:
        initLine()

        for coords in data:
            xlabel, ylabels = initCoords()
            for ylabel in ylabels:
                addData(xlabel, ylabel)
            finishCoords()
        finishLine()

    finishPlot()
    '''
    nlevels = 3

    def __init__(self, *args, **kwargs):
        Renderer.__init__(self, *args, **kwargs)
        Plotter.__init__(self, *args, **kwargs)

    def initPlot(self, fig, work, path):
        '''initialize plot.'''

        self.legend = []
        self.plots = []
        self.xlabels = []
        self.ylabels = []

    def initCoords(self, label, coords):
        '''hook for code working a collection of coords.

        should return a single key for xvalues and
        one or more keys for y-values.
        '''
        keys = list(coords.keys())
        return keys[0], keys[1:]

    def render(self, dataframe, path):

        fig = self.startPlot()

        labels = dataframe.index.levels

        paths = list(itertools.product(*labels))

        self.initPlot(fig, dataframe, path)

        nplotted = 0

        for idx in range(0, len(paths), 2):

            self.initLine(path, dataframe)

            xpath = paths[idx]
            ypath = paths[idx + 1]

            xvalues, yvalues = dataframe.ix[xpath], dataframe.ix[ypath]

            if len(xvalues) != len(yvalues):
                raise ValueError(
                    "length of x,y tuples not consistent: %i != %i" %
                    len(xvalues), len(yvalues))

            R.plot(xvalues, yvalues)
            self.initCoords(xvalues, yvalues)

            nplotted += 1

        self.finishPlot(fig, dataframe, path)

        figid = getCurrentRDevice()
        blocks = ResultBlocks(
            ResultBlock("\n".join(("#$rpl %i$#" % (figid), "")),
                        path2str(path)))

        return blocks


class BoxPlot(Renderer, Plotter):

    """Write a set of box plots.

    This:class:`Renderer` requires two levels.

    labels[dict] / data[array]
    """
    options = Renderer.options + Plotter.options

    nlevels = 1

    def __init__(self, *args, **kwargs):
        Renderer.__init__(self, *args, **kwargs)
        Plotter.__init__(self, *args, **kwargs)

    def render(self, dataframe, path):

        self.startPlot()

        dataseries = Utils.toMultipleSeries(dataframe)

        names = [path2str(x[0]) for x in dataseries]
        data = [x[1] for x in dataseries]
        R.boxplot(data, names=names)

        return self.endPlot(dataframe, path)


class SmoothScatterPlot(Renderer, Plotter):

    """A smoothed scatter plot.

    See R.smoothScatter.

    This:class:`Renderer` requires one levels:

    coords[dict]
    """
    options = Renderer.options + Plotter.options +\
        (('bins', directives.unchanged),)

    nlevels = 2

    def __init__(self, *args, **kwargs):
        Renderer.__init__(self, *args, **kwargs)
        Plotter.__init__(self, *args, **kwargs)

        self.nbins = kwargs.get("nbins", "128")
        if self.nbins:
            if "," in self.nbins:
                self.nbins = list(map(int, self.nbins.split(",")))
            else:
                self.nbins = int(self.nbins)

    def render(self, dataframe, path):

        if len(dataframe.columns) < 2:
            raise ValueError(
                "requiring two coordinates, only got %s" %
                str(dataframe.columns))

        plts, legend = [], []
        blocks = ResultBlocks()

        for xcolumn, ycolumn in itertools.combinations(dataframe.columns, 2):

            # remove missing data points
            xvalues, yvalues = Stats.filterMissing(
                (dataframe[xcolumn], dataframe[ycolumn]))

            # remove columns with all NaN
            if len(xvalues) == 0 or len(yvalues) == 0:
                continue

            # apply log transformation on data not on plot
            if self.logscale:
                if "x" in self.logscale:
                    xvalues = R.log10(xvalues)
                if "y" in self.logscale:
                    yvalues = R.log10(yvalues)

            self.startPlot()
            # wrap, as pandas series can not
            # passed through rpy2.
            R.smoothScatter(numpy.array(xvalues, dtype=numpy.float),
                            numpy.array(yvalues, dtype=numpy.float),
                            xlab=xcolumn,
                            ylab=ycolumn,
                            nbin=self.nbins)
            blocks.extend(self.endPlot(dataframe, path))

        return blocks


class HeatmapPlot(NumpyMatrix, Plotter):

    """A heatmap plot

    See R.heatmap.2 in the gplots package

    """
    options = NumpyMatrix.options + Plotter.options

    nlevels = 2

    def __init__(self, *args, **kwargs):
        NumpyMatrix.__init__(self, *args, **kwargs)
        Plotter.__init__(self, *args, **kwargs)

    def plot(self,  matrix, row_headers, col_headers, path):
        '''plot matrix.

        Large matrices are split into several plots.
        '''

        self.debug("HeatmapPlot started")

        self.startPlot()

        R.library('gplots')

        R["heatmap.2"](matrix,
                       trace='none',
                       dendrogram='none',
                       col=R.bluered(75),
                       symbreaks=True,
                       symkey=True,
                       cexCol=0.5,
                       cexRow=0.5,
                       labRow=row_headers,
                       labCol=col_headers,
                       mar=ro.IntVector((10, 10)),
                       keysize=1)

        self.debug("HeatmapPlot finished")

        return self.endPlot(None, path)

    def render(self, dataframe, path):

        self.debug("building matrix started")
        matrix, rows, columns = self.buildMatrix(dataframe)
        self.debug("building matrix finished")

        return self.plot(matrix, rows, columns, path)


class GGPlot(Renderer, Plotter):

    """Write a set of box plots.

    This:class:`Renderer` requires two levels.

    labels[dict] / data[array]
    """
    options = (
        ('statement',  directives.unchanged),
    ) + Renderer.options + Plotter.options

    nlevels = 1

    def __init__(self, *args, **kwargs):
        Renderer.__init__(self, *args, **kwargs)
        Plotter.__init__(self, *args, **kwargs)

        if "statement" not in kwargs:
            raise ValueError("r-ggplot renderer requires a statement option")

        self.statement = kwargs.get('statement')

    def render(self, dataframe, path):

        R.library('ggplot2')

        # add all indices as columns
        dataframe.reset_index(inplace=True)

        rframe = rpy2.robjects.pandas2ri.py2ri(dataframe)

        # for the issue below, see:
        # http://stackoverflow.com/questions/12865218/getting-rid-of-asis-class-attribute
        unAsIs =  R('''function (x) {
                      if("AsIs" %in% class(x)) {
                          class(x) <- class(x)[-match("AsIs", class(x))]
                      }
                      return (x) } ''')


        rframe = R["as.data.frame"](R.lapply(rframe, unAsIs))
        R.assign("rframe", rframe)

        # start plot
        R('''gp = ggplot(rframe)''')

        # add aesthetics and geometries
        try:
            pp = R('''gp + %s ''' % self.statement)
        except ValueError as msg:
            raise ValueError(
                "could not interprete R statement: "
                "gp + %s; msg=%s" % (self.statement, msg))

        figname = re.sub('/', '_', path2str(path))
        r = ResultBlock('#$ggplot %s$#' % figname,
                        title=path2str(path))
        r.rggplot = pp
        r.figname = figname

        return ResultBlocks(r)



import os
import sys
import re
import json
import StringIO
from docutils.parsers.rst import directives
import numpy
import pandas
# for output of excel work books
import openpyxl

# import numpy.arange and math.log into namepsace that they
# are available in eval
from numpy import arange
from math import log

from CGATReport.ResultBlock import ResultBlock, ResultBlocks
from CGATReport.DataTree import path2str
from CGATReport import Utils, DataTree, Component
from CGATReport import CorrespondenceAnalysis


class Renderer(Component.Component):

    """Base class of renderers that render data into restructured text.

    The subclasses define how to render the data by overloading the
    :meth:`render` method.

    When called, a Renderer and its subclasses will return blocks of
    restructured text. Images are automatically collected from matplotlib
    and inserted at place-holders.

    """
    # plugin fields
    capabilities = ["render"]

    options = (('format', directives.unchanged),
               ('split-at', directives.nonnegative_int),
               ('split-always', directives.unchanged),)

    # required levels in DataTree
    # -1: Renderer can work from any number of levels
    # 0: Renderer needs scalar
    # 1: Renderer requires dataframe with 1 row-label
    # 2: Renderer requires dataframe with 2 row-labels
    # ...
    nlevels = None

    # default number format
    format = "%i"

    # split number of tracks
    split_at = 0

    # tracks always to include if plot is split into
    # several plots.
    split_always = []

    # when splitting, keep first column
    split_keep_first_column = False

    # default group_level for render
    group_level = 0

    # directory where the actual restructured text is located.
    rst_dir = None

    # directory of the root of the source document. rstdir is
    # a subdirectory of this.
    src_dir = None

    # directory in which the document is built.
    build_dir = None

    # dictionary of rst options controlling the display of a
    # directive.
    display_options = {}

    def __init__(self, *args, **kwargs):
        """create an Renderer object using an instance of
        a:class:`Tracker.Tracker`.
        """

        self.debug("%s: starting renderer '%s'" %
                   (id(self), str(self)))

        try:
            self.format = kwargs["format"]
        except KeyError:
            pass

        self.split_at = int(kwargs.get("split-at", 0))

        if "split-always" in kwargs:
            self.split_always = kwargs["split-always"].split(',')

    def __call__(self, dataframe, path):
        '''iterate over leaves/branches in data structure.

        This method will call the:meth:`render` method.

        Large dataframes are split into multiple, smaller rendered
        objects if self.split_at is not zero.

        By default, dataframes are split along the hierachical
        index. However, if there is only a single index, but multiple
        columns, the split is performed on the columns instead. This
        is used when splitting coordinate data as a result of the
        histogram transformation.

        '''
        result = ResultBlocks()

        if not self.split_at:
            result.extend(self.render(dataframe, path))
        else:
            # split dataframe at first index
            level = Utils.getGroupLevels(dataframe)
            grouper = dataframe.groupby(level=level)

            # split dataframe column wise if only one index
            # and multiple columns
            if len(grouper) == 1 and len(dataframe.columns) > self.split_at:
                columns = list(dataframe.columns)
                always = []
                if self.split_keep_first_column:
                    always.append(columns[0])
                # columns to always keep
                always.extend([c for c in columns if c in self.split_always])
                columns = [c for c in columns if c not in always]
                for x in range(0, len(columns), self.split_at):
                    # extract a set of columns
                    result.extend(self.render(
                        dataframe.loc[:, always+columns[x:x+self.split_at]],
                        path))
            # split dataframe along index
            elif len(grouper) >= self.split_at:
                # build groups
                always, remove_always = [], set()

                if self.split_always:
                    for key, work in grouper:
                        for pat in self.split_always:
                            rx = re.compile(pat)
                            if rx.search(path2str(key)):
                                always.append((key, work))
                                remove_always.add(key)

                    grouper = dataframe.groupby(level=level)

                def _group_group(grouper, always, remove_always):
                    group = always[:]
                    for key, work in grouper:

                        if key in remove_always:
                            continue
                        group.append((key, work))

                        if len(group) >= self.split_at:
                            yield group
                            group = always[:]

                    # reconcile index names
                    yield group

                first = True
                for group in _group_group(grouper,
                                          always,
                                          remove_always):
                    # do not plot last dataframe that contains
                    # only the common tracks to plot
                    if not first and len(group) == len(always):
                        continue
                    first = False

                    df = pandas.concat(
                        [x[1] for x in group])

                    # reconcile index names
                    df.index.names = dataframe.index.names
                    result.extend(self.render(df, path))
            else:
                # do not split dataframe
                result.extend(self.render(dataframe, path))

        return result

    def toString(self, value):
        '''returns a number as string

        If not a number, return empty string.'''

        try:
            return self.format % value
        except TypeError:
            return ""

    def set_paths(self, rst_dir, src_dir, build_dir):
        '''set document paths. These are relevant to determine the
        relative position of the object to be rendered in a hierarchy
        of documents.
        '''
        self.rst_dir = rst_dir
        self.src_dir = src_dir
        self.build_dir = build_dir

    def get_paths(self):
        return self.rst_dir, self.src_dir, self.build_dir

    def set_display_options(self, display_options):
        '''set display options given by user.'''
        self.display_options = display_options


class DataFrame(Renderer):
    """
    Renderer outputting the dataframe directly
    as text without any modification.
    """

    options = Renderer.options +\
        (('head', directives.length_or_unitless),
         ('tail', directives.length_or_unitless),
         ('summary', directives.unchanged))

    # output head of dataframe - 0 is no head
    head = 0
    # output tail of dataframe - 0 is no tail
    tail = 0
    # output summary of dataframe
    summary = False

    def __init__(self, *args, **kwargs):

        Renderer.__init__(self, *args, **kwargs)
        self.head = int(kwargs.get('head', 0))
        self.tail = int(kwargs.get('tail', 0))
        self.summary = 'summary' in kwargs

    def __call__(self, dataframe, path):
        result = ResultBlocks()
        texts = []
        if self.head or self.tail:
            if self.head:
                texts.append(str(dataframe.head(self.head)))
            if self.tail:
                texts.append(str(dataframe.tail(self.tail)))
        elif self.summary:
            texts.append(str(dataframe.describe()))
        else:
            texts.append(str(dataframe))

        # add indentation
        texts = ['\n'.join(['   %s' % y for y in x.split('\n')])
                 for x in texts]

        formatted = '''
::

%s

''' % '\n   ...\n'.join(texts)

        result.append(ResultBlock(formatted,
                                  title=path2str(path)))
        return result


class TableBase(Renderer):

    '''base classes for tables and matrices.'''

    options = Renderer.options +\
        (('force', directives.unchanged),
         ('separate', directives.unchanged),
         ('max-rows', directives.length_or_unitless),
         ('large-rows', directives.length_or_unitless),
         ('max-cols', directives.length_or_unitless),
         ('large-html-class', directives.unchanged),
         ('html-class', directives.unchanged))

    max_rows = 50
    max_cols = 20

    def __init__(self, *args, **kwargs):
        Renderer.__init__(self, *args, **kwargs)

        self.force = "force" in kwargs
        self.separate = "separate" in kwargs
        self.max_rows = kwargs.get("max-rows", 50)
        self.max_cols = kwargs.get("max-cols", 20)
        self.html_class = kwargs.get(
            'html-class',
            Utils.PARAMS.get('report_table_class',
                             None))
        self.large_rows = kwargs.get("max-rows", 20)
        self.large_html_class = kwargs.get(
            'large-html-class',
            Utils.PARAMS.get('report_largetable_class',
                             None))

    def asCSV(self, dataframe, row_headers, col_headers, title):
        '''save the table using CSV.'''

        out = StringIO.StringIO()
        dataframe.to_csv(out)
        result = []
        result.append(".. csv-table:: %s" % title)
        lines = out.getvalue().split("\n")

        if len(lines) > self.large_rows:
            if self.large_html_class is not None:
                result.append("   :class: %s" % self.large_html_class)
        else:
            if self.html_class is not None:
                result.append("   :class: %s" % self.html_class)

        if self.add_rowindex:
            raise NotImplementedError('add-rowindex not implemented')
        else:
            result.append('   :header: %s' % lines[0])
            result.append('')
            result.extend(['   %s' % x for x in lines[1:]])

        result.append("")

        return ResultBlock("\n".join(result), title=title)

    def asRST(self, dataframe, row_headers, col_headers, title):
        '''save the table using RST.'''

        out = StringIO.StringIO()
        dataframe.to_csv(out)
        data = [x.split(',') for x in out.getvalue().split('\n')]
        # ignore last element - empty
        del data[-1]

        ncols = len(data[0])
        max_widths = [max([len(row[x]) for row in data])
                      for x in range(ncols)]

        separator = "+" + "+".join(["-" * (x + 2) for x in max_widths]) + "+"
        format = "|" + "|".join([" %%%is " % x for x in max_widths]) + "|"

        lines = []
        lines.append(separator)
        for row in data:
            lines.append(format % tuple(map(str, row)))
            lines.append(separator)
        lines.append("")

        return ResultBlock("\n".join(lines), title=title)

    def asFile(self, dataframe, row_headers, col_headers, title):
        '''save the table as HTML file.

        Multiple files of the same Renderer/Tracker combination are
        distinguished by the title.
        '''

        self.debug("%s: saving %i x %i table as file'" %
                   (id(self),
                    len(row_headers),
                    len(col_headers)))
        lines = []
        lines.append("`%i x %i table <#$html %s$#>`__" %
                     (len(row_headers), len(col_headers),
                      title))

        r = ResultBlock("\n".join(lines) + "\n", title=title)

        out = StringIO.StringIO()
        dataframe.to_csv(out)
        lines = out.getvalue().split("\n")

        # create an html table
        data = ["<table>"]
        data.append("<tr><th></th><th>%s</th></tr>" %
                    "</th><th>".join(map(str, lines[0].split(","))))
        data.extend(["<tr><td>%s</td></tr>" %
                     ("</td><td>".join(x.split(","))) for x in lines[1:]])
        data.append("</table>\n")

        # substitute links
        data = [re.sub(
            "`(.*?(?:\".+\"|\'.+\')?.*?)\s<(.*?(?:\".+\"|\'.+\')?.*?)>`_",
            r'<a href="\2">\1</a>',
            x)
            for x in data]

        r.html = "\n".join(data)

        return r

    def asSpreadSheet(self, dataframe, row_headers, col_headers, title):
        '''save the table as an xls file.

        Multiple files of the same Renderer/Tracker combination are
        distinguished by the title.
        '''

        self.debug("%s: saving %i x %i table as spread-sheet'" %
                   (id(self),
                    len(row_headers),
                    len(col_headers)))

        is_hierarchical = isinstance(dataframe.index,
                                     pandas.core.index.MultiIndex)

        split = is_hierarchical and len(dataframe.index.levels) > 1

        quick = len(dataframe) > 10000
        if quick and not split:
            # quick writing, only append method works
            wb = openpyxl.Workbook(optimized_write=True)

            def fillWorksheet(ws, dataframe, title):
                ws.append([""] + list(col_headers))
                for x, row in enumerate(dataframe.iterrows()):
                    ws.append([path2str(row[0])] + list(row[1]))

                # patch: maximum title length seems to be 31
                ws.title = title[:30]

        else:
            # do it cell-by-cell, this might be slow
            wb = openpyxl.Workbook(optimized_write=False)

            def fillWorksheet(ws, dataframe, title):
                # regex to detect rst hypelinks
                regex_link = re.compile('`(.*) <(.*)>`_')
                # write row names
                for row, row_name in enumerate(dataframe.index):
                    # rows and columns start at 1
                    c = ws.cell(row=row + 2, column=1)
                    c.value = row_name

                # write columns
                for column, column_name in enumerate(dataframe.columns):
                    # set column title
                    # rows and columns start at 1
                    c = ws.cell(row=1, column=column + 2)
                    c.value = column_name

                    # set column values
                    dataseries = dataframe[column_name]

                    if dataseries.dtype == object:
                        for row, value in enumerate(dataseries):
                            c = ws.cell(row=row + 2,
                                        column=column + 2)
                            value = str(value)
                            if value.startswith('`'):
                                c.value, c.hyperlink =\
                                    regex_link.match(value).groups()
                            else:
                                c.value = value
                    else:
                        for row, value in enumerate(dataseries):
                            c = ws.cell(row=row + 2,
                                        column=column + 2)
                            c.value = value
                # patch: maximum title length seems to be 31
                ws.title = re.sub("/", "_", title)[:30]

        if len(wb.worksheets) == 0:
            wb.create_sheet()

        if split:
            # create separate worksheets for nested indices
            nlevels = len(dataframe.index.levels)
            paths = map(tuple, DataTree.unique(
                [x[:nlevels - 1]
                 for x in dataframe.index.unique()]))

            ws = wb.worksheets[0]
            ws.title = 'Summary'
            ws.append(
                [""] * (nlevels - 1) + ["Worksheet", "Rows"])

            for row, path in enumerate(paths):
                # select data frame as cross-section
                work = dataframe.xs(path, axis=0)
                title = path2str(path)
                if len(title) > 30:
                    title = "sheet%i" % row

                ws.append(list(path) + [title, len(work)])
                c = ws.cell(row=row + 1,
                            column=nlevels)
                # this does not work in oocalc
                c.hyperlink = "#%s!A1" % title
                fillWorksheet(wb.create_sheet(),
                              work,
                              title=title)
        else:
            fillWorksheet(wb.worksheets[0], dataframe,
                          title=title)

        # write result block
        lines = []
        lines.append("`%i x %i table <#$xls %s$#>`__" %
                     (len(row_headers), len(col_headers),
                      title))
        lines.append("")

        r = ResultBlock("\n".join(lines), title=title)
        r.xls = wb

        self.debug("%s: saved %i x %i table as spread-sheet'" %
                   (id(self),
                    len(row_headers),
                    len(col_headers)))
        return r


class Table(TableBase):

    '''a basic table.

    Values are either text or converted to text.

    If the has more rows than:attr:`max_rows` or more columns
    than:attr:`max_cols`, a placeholder will be inserted pointing
    towards a file.

    The attribute:attr:`large` determines where large tables are
    written to. The default is html. Alternative values are ``xls``
    for excel spread-sheets.

    '''
    options = TableBase.options +\
        (('transpose', directives.unchanged),
         ('add-rowindex', directives.unchanged),
         ('add-percent', directives.unchanged),
         ('head', directives.length_or_unitless),
         ('large', directives.unchanged),
         ('preview', directives.unchanged))

    # By default, group everything together
    group_level = "all"

    transpose = False

    def __init__(self, *args, **kwargs):
        TableBase.__init__(self, *args, **kwargs)

        self.transpose = "transpose" in kwargs
        self.add_rowindex = "add-rowindex" in kwargs
        self.large = kwargs.get("large", "html")
        self.preview = "preview" in kwargs
        self.add_percent = kwargs.get('add-percent', None)
        self.head = int(kwargs.get('head', 0))

    def modifyTable(self, dataframe):
        '''modify table if required, for example adding percentages.
        '''

        if self.head > 0:
            dataframe = dataframe[:self.head]

        if self.add_percent:
            columns = dataframe.columns
            parts = self.add_percent.split(";")
            for part in parts:
                total, other_col = None, None
                if "," in part:
                    column, method = part.split(",")
                    if method in columns:
                        other_col = columns.index(method)
                    else:
                        try:
                            total = float(method)
                        except ValueError:
                            raise ValueError(
                                ("can't compute total from "
                                 "expression `%s` - missing column?") %
                                method)
                else:
                    column = part

                if column not in columns:
                    raise ValueError("unknown column `%s` to add-percent" %
                                     (column))

                values = numpy.array(dataframe.sum(axis=1),
                                     dtype=numpy.float)

                if other_col:
                    dataframe['%s/%%' % column] = 100.0 *\
                        dataframe[column] / dataframe[other_col]
                else:
                    if total is None:
                        total = float(sum(values))
                    dataframe['%s/%%' % column] = 100.0 *\
                        dataframe[column] / values

        if self.transpose:
            dataframe = dataframe.transpose()
            # flatten the column index if it is hierarchical
            is_hierarchical = isinstance(dataframe.columns,
                                         pandas.core.index.MultiIndex)
            if is_hierarchical:
                dataframe.columns = ["/".join(x) for x in dataframe.columns]

        # if index is not hierarchical, but contains tuples,
        # split tuples in index to build a new (hierarchical) index
        is_hierarchical = isinstance(dataframe.index,
                                     pandas.core.index.MultiIndex)

        if not is_hierarchical and isinstance(dataframe.index[0], tuple):
            idx = pandas.MultiIndex.from_tuples(dataframe.index)
            dataframe.index = idx

        return dataframe

    def __call__(self, dataframe, path):

        # modify table (adding/removing columns) according to user options
        # matrix, row_headers, col_headers = \
        # self.modifyTable(matrix, row_headers, col_headers)
        dataframe = self.modifyTable(dataframe)

        title = path2str(path)

        results = ResultBlocks()

        row_headers = dataframe.index
        col_headers = dataframe.columns

        # as of sphinx 1.3.1, tables with more than 100 columns cause an
        # error:
        # Exception occurred:
        # File "/ifs/apps/apps/python-2.7.9/lib/python2.7/site-packages/docutils/writers/html4css1/__init__.py", line 642, in write_colspecs
        # colwidth = int(node['colwidth'] * 100.0 / width + 0.5)
        # ZeroDivisionError: float division by zero
        #
        # Thus, for table with more than 100 columns, force will be
        # disabled and max_cols set to a low value in order to make
        # sure the table is not displayed inline
        if len(col_headers) >= 90:
            self.force = False
            self.max_cols = 10

        # do not output large matrices as rst files
        if self.separate or (not self.force and
                             (len(row_headers) > self.max_rows or
                              len(col_headers) > self.max_cols)):
            if self.large == "xls":
                results.append(self.asSpreadSheet(dataframe, row_headers,
                                                  col_headers, title))
            else:
                results.append(self.asFile(dataframe, row_headers,
                                           col_headers, title))

            if self.preview:
                raise NotImplementedError('preview not implemented')
                row_headers = row_headers[:self.max_rows]
                col_headers = col_headers[:self.max_cols]
                # matrix = [x[:self.max_cols] for x in
                #          matrix[:self.max_rows]]
            else:
                return results

        results.append(self.asCSV(dataframe, row_headers, col_headers, title))

        return results


class HTMLTable(Table):

    '''a table in html format to download.
    '''

    def __call__(self, dataframe, path):

        results = ResultBlocks()
        if dataframe is None:
            return results

        title = path2str(path)
        row_headers = dataframe.index
        col_headers = dataframe.columns
        results = ResultBlocks()
        results.append(self.asFile(dataframe, row_headers,
                                   col_headers, title))

        return results


class XlsTable(Table):

    '''a table in xls format to download.
    '''

    def __call__(self, dataframe, path):

        results = ResultBlocks()
        if dataframe is None:
            return results

        title = path2str(path)
        row_headers = dataframe.index
        col_headers = dataframe.columns
        results.append(self.asSpreadSheet(dataframe, row_headers,
                                          col_headers, title))

        return results


class RstTable(Table):

    '''an rst formatted table.
    '''

    def __call__(self, dataframe, path):

        results = ResultBlocks()

        if dataframe is None:
            return results

        title = path2str(path)

        row_headers = dataframe.index
        col_headers = dataframe.columns

        # do not output large matrices as rst files
        if self.separate or (not self.force and
                             (len(row_headers) > self.max_rows or
                              len(col_headers) > self.max_cols)):
            if self.large == "xls":
                results.append(self.asSpreadSheet(dataframe, row_headers,
                                                  col_headers, title))
            else:
                results.append(self.asFile(dataframe, row_headers,
                                           col_headers, title))

        results.append(self.asRST(dataframe, row_headers, col_headers, title))

        return results


class GlossaryTable(Table):

    """output a table in the form of a glossary."""

    def __call__(self, dataframe, path):

        results = ResultBlocks()

        if dataframe is None:
            return results

        title = path2str(path)

        lines = []
        lines.append(".. glossary::")
        lines.append("")

        for x, row in enumerate(dataframe.iterrows()):
            header, data = row
            txt = "\n      ".join([x.strip() for x in str(data).split("\n")])
            lines.append('   %s\n      %s\n' % (path2str(header), txt))

        lines.append("")

        results.append(ResultBlock("\n".join(lines), title=title))

        return results


class MatrixBase:

    '''base class for matrices.

    This base class provides utility functions for rectangular 2D matrices.

    It implements column-wise and row-wise transformations.

    This class adds the following options to the:term:`report` directive.

    :term:`transform-matrix`: apply a matrix transform. Possible choices are:

           * *correspondence-analysis*: use correspondence analysis
             to permute rows/columns
           * *normalized-col-total*: normalize by column total
           * *normalized-col-max*: normalize by column maximum,
           * *normalized-col-first*: normalize by first column.
             The first column is then removed.
           * *normalized-row-total*: normalize by row total
           * *normalized-row-max*: normalize by row maximum
           * *normalized-row-first*: normalize by first row.
             The first row is then removed.
           * *normalized-total*: normalize over whole matrix
           * *normalized-max*: normalize over whole matrix
           * *sort*: sort matrix rows and columns alphanumerically.
           * filter-by-rows: only take columns that are also present in rows
           * filter-by-cols: only take columns that are also present in cols
           * square: make square matrix (only take rows and columns present
             in both)
           * *add-row-total*: add the row total at the bottom
           * *add-column-total*: add the column total as a last column

    Requires two levels:

       rows[dict] / columns[dict] / value

    All values need to be numerical.
    '''

    def __init__(self, *args, **kwargs):

        self.mMapKeywordToTransform = {
            "correspondence-analysis": self.transformCorrespondenceAnalysis,
            "transpose": self.transformTranspose,
            "normalized-row-total": self.transformNormalizeRowTotal,
            "normalized-row-max": self.transformNormalizeRowMax,
            "normalized-row-first": self.transformNormalizeRowFirst,
            "normalized-col-total": self.transformNormalizeColumnTotal,
            "normalized-col-max": self.transformNormalizeColumnMax,
            "normalized-col-first": self.transformNormalizeColumnFirst,
            "normalized-total": self.transformNormalizeTotal,
            "normalized-max": self.transformNormalizeMax,
            "symmetric-max": self.transformSymmetricMax,
            "symmetric-min": self.transformSymmetricMin,
            "symmetric-avg": self.transformSymmetricAverage,
            "symmetric-sum": self.transformSymmetricSum,
            "filter-by-rows": self.transformFilterByRows,
            "filter-by-cols": self.transformFilterByColumns,
            "square": self.transformSquare,
            "add-row-total": self.transformAddRowTotal,
            "add-column-total": self.transformAddColumnTotal,
            "sort": self.transformSort,
        }
        self.tofloat = False
        self.converters = []
        if "transform-matrix" in kwargs:
            for kw in [x.strip() for x in
                       kwargs["transform-matrix"].split(",")]:
                if kw.startswith("normalized"):
                    self.format = "%6.4f"
                    self.tofloat = True
                try:
                    self.converters.append(self.mMapKeywordToTransform[kw])
                except KeyError:
                    raise ValueError(("unknown matrix transformation %s, "
                                      "possible values are: %s") %
                                     (kw,
                                      ",".join(sorted(
                                          self.mMapKeywordToTransform.keys()
                                      ))))

    def transformAddRowTotal(self, matrix, row_headers, col_headers):
        '''add row total to the matrix.'''
        nrows, ncols = matrix.shape

        totals = numpy.zeros(nrows)
        for x in range(nrows):
            totals[x] = sum(matrix[x, :])
        col_headers.append("total")
        totals.shape = (nrows, 1)
        return (numpy.hstack(numpy.hsplit(matrix, ncols) + [totals, ]),
                row_headers, col_headers)

    def transformAddColumnTotal(self, matrix, row_headers, col_headers):
        raise NotImplementedError

    def transformFilterByRows(self, matrix, row_headers, col_headers):
        """only take columns that are also present in rows"""
        take = [x for x in range(len(col_headers))
                if col_headers[x] in row_headers]
        return (matrix.take(take, axis=1), row_headers,
                [col_headers[x] for x in take])

    def transformFilterByColumns(self, matrix, row_headers, col_headers):
        """only take rows that are also present in columns"""
        take = [x for x in range(len(row_headers))
                if row_headers[x] in col_headers]
        return (matrix.take(take, axis=0),
                [row_headers[x] for x in take],
                col_headers)

    def transformSquare(self, matrix, row_headers, col_headers):
        """only take rows and columns that are present in both giving a square
        matrix.
        """
        take = set(row_headers).intersection(set(col_headers))
        row_indices = [x for x in range(len(row_headers))
                       if row_headers[x] in take]
        col_indices = [x for x in range(len(col_headers))
                       if col_headers[x] in take]
        m1 = matrix.take(row_indices, axis=0)
        m2 = m1.take(col_indices, axis=1)

        return (m2,
                [row_headers[x] for x in row_indices],
                [col_headers[x] for x in col_indices])

    def transformTranspose(self, matrix, row_headers, col_headers):
        """transpose the matrix."""
        return numpy.transpose(matrix), col_headers, row_headers

    def transformCorrespondenceAnalysis(self, matrix,
                                        row_headers,
                                        col_headers):
        """apply correspondence analysis to a matrix.
        """
        if len(row_headers) <= 1 or len(col_headers) <= 1:
            self.warn("correspondence analysis skipped for "
                      "matrices with single row/column")
            return matrix, row_headers, col_headers

        try:
            row_indices, col_indices = \
                CorrespondenceAnalysis.GetIndices(matrix)
        except ValueError:
            return matrix, row_headers, col_headers

        map_row_new2old = numpy.argsort(row_indices)
        map_col_new2old = numpy.argsort(col_indices)

        matrix, row_headers, col_headers = \
            CorrespondenceAnalysis.GetPermutatedMatrix(matrix,
                                                       map_row_new2old,
                                                       map_col_new2old,
                                                       row_headers=row_headers,
                                                       col_headers=col_headers)

        return matrix, row_headers, col_headers

    def transformSort(self, matrix, row_headers, col_headers):
        """apply correspondence analysis to a matrix.
        """

        map_row_new2old = [x[0] for x in sorted(enumerate(row_headers),
                                                key=lambda x: x[1])]
        map_col_new2old = [x[0] for x in sorted(enumerate(col_headers),
                                                key=lambda x: x[1])]

        matrix, row_headers, col_headers = \
            CorrespondenceAnalysis.GetPermutatedMatrix(matrix,
                                                       map_row_new2old,
                                                       map_col_new2old,
                                                       row_headers=row_headers,
                                                       col_headers=col_headers)

        return matrix, row_headers, col_headers

    def transformSymmetricMax(self, matrix, rows, cols):
        """symmetrize a matrix.

        returns the normalized matrix.
        """
        if len(rows) != len(cols):
            raise ValueError("matrix not square - can not be symmetrized")

        for x in range(len(rows)):
            for y in range(x + 1, len(cols)):
                matrix[x, y] = matrix[y, x] = max(matrix[x, y], matrix[y, x])

        return matrix, rows, cols

    def transformSymmetricMin(self, matrix, rows, cols):
        """symmetrize a matrix.

        returns the normalized matrix.
        """
        if len(rows) != len(cols):
            raise ValueError("matrix not square - can not be symmetrized")

        for x in range(len(rows)):
            for y in range(x + 1, len(cols)):
                matrix[x, y] = matrix[y, x] = \
                    min(matrix[x, y], matrix[y, x])

        return matrix, rows, cols

    def transformSymmetricSum(self, matrix, rows, cols):
        """symmetrize a matrix.

        returns the normalized matrix.
        """
        if len(rows) != len(cols):
            raise ValueError("matrix not square - can not be symmetrized")

        for x in range(len(rows)):
            for y in range(x + 1, len(cols)):
                matrix[x, y] = matrix[y, x] = \
                    sum(matrix[x, y], matrix[y, x])

        return matrix, rows, cols

    def transformSymmetricAverage(self, matrix, rows, cols):
        """symmetrize a matrix.

        returns the normalized matrix.
        """
        if len(rows) != len(cols):
            raise ValueError("matrix not square - can not be symmetrized")

        for x in range(len(rows)):
            for y in range(x + 1, len(cols)):
                matrix[x, y] = matrix[y, x] = \
                    sum(matrix[x, y], matrix[y, x]) / 2

        return matrix, rows, cols

    def transformNormalizeTotal(self, matrix, rows, cols):
        """normalize a matrix by the total.

        Returns the normalized matrix.
        """
        return matrix / matrix.sum(), rows, cols

    def transformNormalizeMax(self, matrix, rows, cols):
        """normalize a matrix by the max.

        Returns the normalized matrix.
        """
        return matrix / matrix.max(), rows, cols

    def transformNormalizeRowTotal(self, matrix, rows, cols):
        """normalize a matrix row by the row total.

        Returns the normalized matrix.
        """
        return matrix / matrix.sum(axis=1), rows, cols

    def transformNormalizeRowMax(self, matrix, rows, cols):
        """normalize a matrix row by the row maximum.

        Returns the normalized matrix.
        """
        return matrix / matrix.max(axis=1), rows, cols

    def transformNormalizeRowFirst(self, matrix, rows, cols):
        """normalize a matrix row by the row maximum.

        Returns the normalized matrix.
        """
        m = matrix[0]
        matrix = numpy.delete(matrix, 0, 0)
        return matrix / m, rows[1:], cols

    def transformNormalizeColumnTotal(self, matrix, rows, cols):
        """normalize a matrix by the column total.

        Returns the normalized matrix."""
        return matrix / matrix.sum(axis=0), rows, cols

    def transformNormalizeColumnFirst(self, matrix, rows, cols):
        """normalize a matrix by the first column.

        Removes the first column.

        Returns the normalized matrix."""
        m = matrix[:, 0]
        matrix = numpy.delete(matrix, 0, 1)
        return matrix / m, rows, cols[1:]

    def transformNormalizeColumnMax(self, matrix, rows, cols):
        """normalize a matrix by the column maximum

        Returns the normalized matrix.
        """
        return matrix / matrix.max(axis=0), rows, cols

    def render(self, work, path):
        """render the data.
        """

        results = ResultBlocks(title=path)

        matrix, rows, columns = self.buildMatrix(work)

        title = path2str(path)

        if len(rows) == 0:
            return ResultBlocks(ResultBlock("", title=title))

        # do not output large matrices as rst files
        # separate and force need to be mixed in.
        if self.separate or (not self.force and
                             (len(rows) > self.max_rows or
                              len(columns) > self.max_cols)):
            return ResultBlocks(self.asFile(pandas.DataFrame(matrix,
                                                             index=rows,
                                                             columns=columns),
                                            rows,
                                            columns,
                                            title),
                                title=path)

        lines = []
        lines.append(".. csv-table:: %s" % title)
        lines.append('   :header: "track","%s" ' % '","'.join(columns))
        lines.append('')
        for row in range(len(rows)):
            lines.append(
                '   "%s","%s"' %
                (rows[row], '","'.join(
                    [self.toString(x) for x in matrix[row]])))
        lines.append("")

        if path is None:
            subtitle = ""
        else:
            subtitle = path2str(path)

        results.append(ResultBlock("\n".join(lines), title=subtitle))

        return results


class TableMatrix(TableBase, MatrixBase):

    """A table with numerical columns.

       rows[dict] / columns[dict] / value

    All values need to be numerical.
    """

    # By default, group everything together
    group_level = "all"

    options = TableBase.options +\
        (('transform-matrix', directives.unchanged),)

    # wether or not to convert matrix to float
    tofloat = False

    def __init__(self, *args, **kwargs):

        TableBase.__init__(self, *args, **kwargs)
        MatrixBase.__init__(self, *args, **kwargs)

    def buildMatrix(self,
                    dataframe,
                    missing_value=0,
                    apply_transformations=True,
                    dtype=numpy.float):
        """build a matrix from work, a two-level nested dictionary.

        This method will also apply conversions if apply_transformations
        is set.
        """
        rows = map(path2str, dataframe.index)
        columns = list(dataframe.columns)
        # use numpy.matrix - permits easier broadcasting
        # for normalization.
        matrix = numpy.matrix(dataframe.as_matrix())

        # remove columns with only NaNs. This can happend
        # during the dataframe merging process if the
        # columns are unique to each matrix.
        take = numpy.array(numpy.all(numpy.isnan(matrix), axis=0).flat)
        matrix = matrix[:, ~take]
        columns = list(numpy.array(columns)[~take])

        if self.converters and apply_transformations:
            # convert to float for conversions
            if self.tofloat:
                matrix = numpy.matrix(matrix, dtype=numpy.float)

            for converter in self.converters:
                self.debug("applying converter %s" % converter)
                matrix, rows, columns = converter(matrix, rows, columns)

        # convert rows/columns to str (might be None)
        rows = [str(x) for x in rows]
        columns = [str(x) for x in columns]

        # convert numpy.matrix to numpy.array
        return numpy.asarray(matrix), rows, columns

# for compatibility
Matrix = TableMatrix


class NumpyMatrix(TableMatrix, MatrixBase):

    """Deprecated - not needed any more as equivalent to TableMatrix
    """

    def __init__(self, *args, **kwargs):
        TableMatrix.__init__(self, *args, **kwargs)
        MatrixBase.__init__(self, *args, **kwargs)


class Status(Renderer):

    '''Renders a status report.

    A status report is a two element table containing
    status ('PASS', 'FAIL', 'WARNING', 'NA') and some information.

    The __doc__ string of the tracker is added as a legend below
    the status report.
    '''

    options = Renderer.options +\
        (('no-legend', directives.unchanged),)


    # read complete data
    nlevels = -1

    # For backwards compatibility, group by slice (test function)
    # and not track (data track) by default.
    group_level = 'slice'

    map_code2image = {'FAIL': "fail.png",
                      'PASS': "pass.png",
                      'NA': "not_available.png",
                      'WARNING': "warning.png",
                      'WARN': "warning.png"}

    display_legend = True

    def __init__(self, *args, **kwargs):
        
        Renderer.__init__(self, *args, **kwargs)
        if "no-legend" in kwargs:
            self.display_legend = False

    def __call__(self, dataframe, path):

        # convert to dataframe
        # index has test names
        # columns are description, info, status
        columns = ('description', 'info', 'status', 'name')
        if set(dataframe.columns) != set(columns):
            raise ValueError("invalid columns: expected '%s', got '%s' " %
                             (columns, dataframe.columns))

        lines = []
        dirname = os.path.join(os.path.dirname(
            sys.modules["CGATReport"].__file__), "images")
        descriptions = {}
        title = "status"

        # add header
        lines.append(".. csv-table:: %s" % "table")
        lines.append('   :header: "Track", "Test", "", "Status", "Info"')
        lines.append('')

        for index, values in dataframe.iterrows():
            testname = values['name']
            description = values['description']
            info = values['info']
            status = values['status']
            track = path2str(index)

            descriptions[testname] = description

            try:
                image = ".. image:: %s" %\
                    os.path.join(dirname, self.map_code2image[status.upper()])
            except KeyError:
                image = ""

            lines.append(
                '   "%(track)s", ":term:`%(testname)s`", "%(image)s", "%(status)s", "%(info)s"' %
                locals())

        lines.append("")

        if self.display_legend:
            lines.append(".. glossary::")
            lines.append("")

            for test, description in descriptions.items():
                lines.append('%s\n%s\n' % (Utils.indent(test, 3),
                                           Utils.indent(description, 6)))

        return ResultBlocks(ResultBlock("\n".join(lines), title=""))


class DataTreeRenderer(object):
    """Base class for renderers processing a data tree
    directly.

    Note that these are special case renderers.
    """
    # plugin fields
    capabilities = ["render"]

    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, data):
        return self.render(data)


class User(DataTreeRenderer):

    """Renderer for user-implemented rendering.

    The renderer itself creates no output, but returns the results
    of the tracker.

    When called, a Renderer and its subclasses will return blocks of
    restructured text. Images are automatically collected from matplotlib
    and other renderers from active graphics devices and inserted at
    the place-holders.
    """

    def __init__(self, *args, **kwargs):
        DataTreeRenderer.__init__(self, *args, **kwargs)

    def render(self, data):

        # initiate output structure
        results = ResultBlocks(title='user')

        labels = DataTree.getPaths(data)
        # iterate over all items at leaf
        for path, branch in DataTree.getNodes(data, len(labels) - 2):
            for key in Utils.TrackerKeywords:
                if key in branch:
                    # add a result block
                    results.append(ResultBlock(branch[key],
                                               title=path2str(path)))

        return results


class Debug(DataTreeRenderer):

    '''a simple renderer, returning the type of data
    and the number of items at each path.'''

    def render(self, data):

        # initiate output structure
        results = ResultBlocks(title='debug')

        try:
            results.append(ResultBlock(json.dumps(
                data, indent=4), title=''))
        except TypeError:
            results.append(ResultBlock(str(data), title=''))

        return results



from CGATReport.Tracker import *
#from CGATReportGeneLists.Tracker import *


class SimpleExampleData(Tracker):

    """Simple Example Data.
    """

    tracks = ["bicycle", "car"]

    def __call__(self, track, slice=None):
        if track == "car":
            return odict((("wheels", 4), ("max passengers", 5)))
        elif track == "bicycle":
            return odict((("wheels", 2), ("max passengers", 1)))


class DifferentialHead(TrackerSQL):

    def __call__(self, track):

        statement = "SELECT * FROM differential LIMIT 10"
        return self.getAll(statement)


class GeneListHead(TrackerSQL):

    def __call__(self, track):

        statement = "SELECT gene_id FROM Promoters_with_ar LIMIT 10"
        return {'gene_id': self.getValues(statement)}


class SimpleOverlap(TrackerMultipleLists):

    statements = {"AR": "SELECT gene_id FROM Promoters_with_ar",
                  "ERG": "SELECT gene_id FROM Promoters_with_erg",
                  "background": "SELECT gene_id FROM all_genes"}


class OverlapTracker(TrackerMultipleLists):
    pattern = "(.+)_with_.+"
    slices = ["logFC < 0", "logFC > 0"]

    # Note that the column names in the lists must match
    ListA = '''SELECT id as gene_id
                   FROM differential
                   WHERE FDR < 0.05 AND %(slice)s '''

    ListB = '''SELECT gene_id
    FROM %(track)s_with_ar as ar,
    differential as diff
    WHERE ar.gene_id = diff.id'''

    ListC = '''SELECT gene_id
                   FROM %(track)s_with_erg as erg,
                         differential as diff
                   WHERE erg.gene_id = diff.id'''

    background = '''SELECT id AS gene_id FROM differential'''

    # we also need to add backround to the labels
    labels = ["Differentially Expressed",
              "Bound by AR",
              "Bound by ERG",
              "background"]



'''This document contains a number of sample trackers.'''

import sys, os, re, random, glob

import numpy

from SphinxReport.Tracker import Tracker, Status
from collections import OrderedDict as odict

def BarData(): return dict( [("bar1", 20), ("bar2", 10)] )

class LabeledDataExample( Tracker ):
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")
    def __call__(self, track, slice = None):
        if slice == "slice1":
            return odict( (("column1", 10),
                           ("column2", 20 ),) )
        elif slice == "slice2":
            return odict ( (("column1", 20),
                            ("column2", 10),
                            ("column3", 5),) )

class LabeledDataLargeExample( Tracker ):
    slices = ("slice1", "slice2")
    tracks = [ "track%i" % x for x in range(0,100) ]
    def __call__(self, track, slice = None):
        if slice == "slice1":
            return odict( (("column1", 10),
                           ("column2", 20 ),) )
        elif slice == "slice2":
            return odict ( (("column1", 20),
                            ("column2", 10),
                            ("column3", 5),) )

class LabeledDataWithErrorsExample( Tracker ):
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")
    def __call__(self, track, slice = None):
        if slice == "slice1":
            return odict( ( 
                    ("column1", odict( ( ('data', 20), 
                                         ('error', 1), 
                                         ) ) ),
                    ("column2", odict( ( ('data', 10), 
                                         ('error', 2),
                                         ) ) ),
                    )) 
        elif slice == "slice2":
            return odict( ( 
                    ("column1", odict( ( ('data', 20), 
                                         ('error', 3),
                                         ) )),
                    ("column2", odict( ( ('data', 10), 
                                         ('error', 4))) ),
                    ("column3", odict( ( ('data', 30), 
                                         ('error', 5))) ),
                    ) )

class LabeledDataWithErrorsAndLabelsExample( Tracker ):
    '''use of ordered dictionary important as first
    element is treated as data.
    '''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")
    def __call__(self, track, slice = None):
        if slice == "slice1":
            return odict( ( 
                    ("column1", odict( ( ('data', 20), 
                                         ('error',5), 
                                         ('label','**' ) ) )),
                    ("column2", odict( ( ('data', 10), 
                                         ('error',2), 
                                         ('label', '*' ) ) ) )
                    )) 
        elif slice == "slice2":
            return odict( ( 
                    ("column1", odict( ( ('data', 20), 
                                         ('error',5),
                                         ('label','***' ) ) )),
                    ("column2", odict( ( ('data', 10), 
                                         ('error',1))) ),
                    ("column3", odict( ( ('data', 30), 
                                         ('error',4))) ),
                    ) )

class SingleColumnDataExample( Tracker ):
    '''return a single column of data.'''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")
    def __call__(self, track, slice = None):
        s = [random.randint(0,20) for x in range(40)]
        random.shuffle( s )
        return odict( (("data", s),) )

class ArrayDataExample( Tracker ):
    '''return two arrays of data.'''
    
    slices = [ "slice%i" % x for x in range(0,2) ]
    tracks = [ "track%i" % x for x in range(0,3) ]

    def __call__(self, track, slice = None):

        scale = (3-int(track[-1])) 
        
        data = odict( (("x", range(0,50)),
                       ("y", [ x * scale for x in range(0,50) ] ) ) )
        
        return data

class SingleColumnDataLargeExample( Tracker ):
    '''return a single column of data.'''
    slices = ("slice1", "slice2")
    tracks = [ "track%i" % x for x in range(0,20) ]
    def __call__(self, track, slice = None):
        s = [random.randint(0,20) for x in range(40)]
        random.shuffle( s )
        return odict( (("data", s),) )


class SingleColumnDataWithErrorExample( Tracker ):
    '''return a single column of data.'''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")
    def __call__(self, track, slice = None):
        s = [random.randint(0,20) for x in range(40)]
        e = [random.randint(0,3) for x in range(40)]
        random.shuffle( s )
        return odict( (("data", s),
                       ("error", e) ) )

class SingleColumnDataExampleWithoutSlices( Tracker ):
    '''return a single column of data.'''
    tracks = ("track1", "track2", "track3")
    def __call__(self, track, slice = None):
        s = [random.randint(0,20) for x in range(40)]
        random.shuffle( s )
        return odict( (("data", s),) )

class MultipleColumnDataExample( Tracker ):
    '''multiple columns each with a column with data.'''
    mColumns = [ "col1", "col2", "col3" ]
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2")
    def __call__(self, track, slice = None):
        data = []
        if slice == "slice1":
            for x in range(len(self.mColumns)-1):
                data.append( [ y + random.gauss( 0, 0.2 ) for y in range(20) ] )
        elif slice == "slice2":
            for x in range(len(self.mColumns)):
                data.append( [ y + random.gauss( 0, 0.5 ) for y in range(20) ] )
        return odict( zip(self.mColumns, data) )

class MultipleColumnDataFullExample( Tracker ):
    '''multiple columns each with a column with data.'''
    mColumns = [ "col1", "col2", "col3" ]
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2")
    def __call__(self, track, slice = None):
        data = []
        if slice == "slice1":
            for x in range(len(self.mColumns)):
                data.append( [ y + random.gauss( 0, 0.2 ) for y in range(20) ] )
        elif slice == "slice2":
            for x in range(len(self.mColumns)):
                data.append( [ y + random.gauss( 0, 0.5 ) for y in range(20) ] )
        return odict( zip(self.mColumns, data) )

class ErrorInTracker1( Tracker ):
    '''A tracker that creates an error - exception while collecting data.'''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2")
    def __call__(self, track, slice = None):
        raise ValueError("testing: could not collect data")
    
class ErrorInTracker2( Tracker ):
    '''A tracker that creates an error - problems while returning tracks.'''
    slices = ("slice1", "slice2")
    def getTracks( self ):
        raise ValueError( "testing: could not build tracks" )
    def __call__(self, track, slice = None):
        return odict( (("data", range(0,10)),) )

class ErrorInTracker3( Tracker ):
    '''A tracker that creates an error - no data.'''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2")
    def __call__(self, track, slice = None):
        return None

class EmptyTracker( Tracker ):
    '''A tracker that creates a warning.'''
    slices = ("slice1", "slice2")
    tracks = []
    def __call__(self, track, slice = None):
        return None

class LabeledDataTest( Tracker ):
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice = None):
        if slice == "slice1":
            return odict( (("column1", 10),
                           ("column2", 20 ),) )
        elif slice == "slice2":
            return odict ( (("column1", 20),
                            ("column2", 10),
                            ("column3", 5),) )
    
class StatusTracker( Status ):
    tracks = ("track1", "track2", "track3")
    
    def testTest1( self, track ):
        '''test1 passes'''
        return "PASS", 0.5

    def testTest2( self, track ):
        '''test2 fails - 
        A large test.'''
        return "FAIL", 2

    def testTest3( self, track ):
        '''test3 gives a warning'''
        return "WARN", "a string"

    def testTest4( self, track ):
        '''test4 is not available/applicable'''
        return "NA", None


class MatrixTracker( Tracker ):
    '''returns matrices.'''

    tracks = 'example1', 'example2'
    
    def __call__(self, track ):
        
        if track == 'example1':
            matrix =  numpy.arange( 0, 10 )
        elif track == 'example2':
            matrix = numpy.arange( 10,0,-1 )
            
        matrix.shape = (2,5)

        r = { 'rows' : map(str, xrange( 0, 2 ) ),
              'columns' : map(str, xrange( 0, 5 ) ),
              'matrix' : matrix  }

        return r

def getSingleValue(): return 12
    
IMAGEDIR = os.path.join( os.path.dirname( os.path.abspath( __file__ )), "..", "images")

class DataWithImagesExample( Tracker ):
    tracks = ("all",)

    def __call__(self, track):

        images = glob.glob( os.path.join( IMAGEDIR, "*.png") )

        data = odict()
        data["numbers"] = range( len(images))
        data["images"] = [".. image:: %s"  % x for x in images ]
        
        return data


class VennTracker( Tracker ):
    
    tracks = ('two-circle', 'three-circle' )

    def __call__( self, track ):

        if track == 'two-circle':
            return {'01' : 10, '10': 20, '11' : 5, 
                    'labels' : ("SetA", "SetB") }

        elif track == 'three-circle':
            return {'001' : 10, '010': 20, '100' : 5,
                    '011' : 10, '110': 20, '101' : 5,
                    '111' : 10,
                    'labels' : ("SetA", "SetB", "SetC") }




from docutils.parsers.rst import directives

from Plotter import TableMatrixPlot, DataSeriesPlot
import pandas
import seaborn
import numpy
from CGATReportPlugins.Renderer import Renderer
from CGATReportPlugins.Plotter import Plotter


class SeabornPlot(object):

    options =  (('kwargs', directives.unchanged),)
    
    kwargs = ""

    def __init__(self, *args, **kwargs):
        self.kwargs = kwargs.get("kwargs", "")

    def execute(self, statement, g, l):
        """execute seaborn statement. Inserts kwargs."""

        statement = statement.strip()
        assert statement.endswith(")")
        if self.kwargs:
            statement = statement[:-1] + ", %s)" % self.kwargs

        plot = None
        try:
            exec "plot = %s" % statement in g, l
        except Exception, msg:
            raise Exception(
                "seaborn raised error for statement '%s': msg=%s" %
                (statement, msg))
        return plot


class PairPlot(Renderer, Plotter, SeabornPlot):

    def __init__(self, *args, **kwargs):
        Renderer.__init__(self, *args, **kwargs)
        Plotter.__init__(self, *args, **kwargs)
        SeabornPlot.__init__(self, *args, **kwargs)
        
    def render(self, dataframe, path):

        statement = ("seaborn.pairplot(dataframe)")
        plot = self.execute(statement, globals(), locals())

        return self.endPlot([plot], [], path)

class BoxPlot(DataSeriesPlot, SeabornPlot):

    """Write a set of box plots.

    This:class:`Renderer` requires two levels.

    labels[dict] / data[array]
    """

    def __init__(self, *args, **kwargs):
        DataSeriesPlot.__init__(self, *args, **kwargs)
        SeabornPlot.__init__(self, *args, **kwargs)

    def plotData(self, data, melted):
        if melted:
            return [seaborn.boxplot(data.value,
                                    groupby=data.label)]
        else:
            return [seaborn.boxplot(data)]


class ViolinPlot(BoxPlot):

    """Write a set of violin plots.

    This:class:`Renderer` requires two levels.

    labels[dict] / data[array]
    """

    def __init__(self, *args, **kwargs):
        DataSeriesPlot.__init__(self, *args, **kwargs)
        BoxPlot.__init__(self, *args, **kwargs)

    def plotData(self, data, melted):
        if melted:
            return [seaborn.violinplot(data.value,
                                       groupby=data.label)]
        else:
            return [seaborn.violinplot(data)]


class KdePlot(DataSeriesPlot, SeabornPlot):

    """Write a set of density plots.

    The data series is converted to a kernel density
    estimate.

    This:class:`Renderer` requires two levels.

    labels[dict] / data[array]
    """
    options = DataSeriesPlot.options +\
        (('shade', directives.flag),
         ('vertical', directives.flag),
         ('kernel', directives.unchanged),
         ('bw', directives.unchanged),
         ('gridsize', directives.unchanged),
         ('cut', directives.unchanged),
         ('clip', directives.unchanged),
         )

    def __init__(self, *args, **kwargs):
        DataSeriesPlot.__init__(self, *args, **kwargs)
        SeabornPlot.__init__(self, *args, **kwargs)

        self.shade = kwargs.get('shade', False)
        self.vertical = kwargs.get('shade', False)
        self.kernel = kwargs.get('kernel', 'gau')
        self.bw = kwargs.get('bw', 'scott')
        self.gridsize = int(kwargs.get('gridsize', 100))
        self.cut = int(kwargs.get('cut', 3))
        if 'clip' in kwargs:
            self.clip = map(int, kwargs['clip'].split(','))

    def plotData(self, dataframe, melted):
        plts = []
        if melted:
            for key, group in dataframe.groupby(dataframe.label):
                plts.append(seaborn.kdeplot(
                    numpy.array(dataframe.value, dtype=numpy.float),
                    label=key,
                    shade=self.shade,
                    vertical=self.vertical,
                    kernel=self.kernel))
        else:
            for column in dataframe.columns:
                plts.append(seaborn.kdeplot(
                    numpy.array(dataframe[column], dtype=numpy.float),
                    label=column,
                    shade=self.shade,
                    vertical=self.vertical,
                    kernel=self.kernel))
        return plts


class DistPlot(DataSeriesPlot, SeabornPlot):

    """Write a set of density plots.

    The data series is converted to a kernel density
    estimate.

    This:class:`Renderer` requires two levels.

    labels[dict] / data[array]
    """

    def __init__(self, *args, **kwargs):
        DataSeriesPlot.__init__(self, *args, **kwargs)
        SeabornPlot.__init__(self, *args, **kwargs)

    def plotData(self, dataframe, melted):
        plts = []
        if melted:
            for key, group in dataframe.groupby(dataframe.label):
                plot = self.execute(
                    "seaborn.distplot( "
                    "numpy.array(dataframe.value, dtype=numpy.float), "
                    "label=key)", globals(), locals())
                plts.append(plot)
        else:
            for column in dataframe.columns:
                plot = self.execute(
                    "seaborn.distplot( "
                    "numpy.array(dataframe[column], dtype=numpy.float), "
                    "label=column)", globals(), locals())
                plts.append(plot)
        return plts


class SeabornMatrixPlot(TableMatrixPlot, SeabornPlot):
    
    def __init__(self, *args, **kwargs):
        TableMatrixPlot.__init__(self, *args, **kwargs)
        SeabornPlot.__init__(self, *args, **kwargs)

    def addColourBar(self):
        pass
        
    def plotMatrix(self, matrix, row_headers, col_headers,
                   vmin, vmax,
                   color_scheme=None):

        self.debug("plot matrix started")
        data = pandas.DataFrame(matrix,
                                columns=col_headers,
                                index=row_headers)
        
        statement = self.buildStatement(data, color_scheme)
        self.debug("seaborn: executing statement '%s'" % statement)
        plot = self.execute(statement, globals(), locals())
        self.debug("plot matrix finished")
        return plot


class HeatmapPlot(SeabornMatrixPlot):
    """Render a matrix as a heatmap using seaborn.
    """

    def buildStatement(self, data, color_scheme):
        return("plot = seaborn.heatmap(data, "
               "cmap=color_scheme)")


class ClustermapPlot(SeabornMatrixPlot):
    """Render a matrix as a heatmap using seaborn.
    """
    def buildStatement(self, data, color_scheme):

        if data.isnull().any().any():
            raise ValueError("dataframe contains NaN")
        
        return("plot = seaborn.clustermap(data, "
               "cmap=color_scheme)")





from .Logger import warn, debug, info
from . import Logger
import multiprocessing

Logger.basicConfig(
    level=Logger.DEBUG,
    format='%(asctime)s %(levelname)s %(message)s',
    filename="logger_test.log")


def test_logging(args):
    msg = args[0]
    info(msg)

if __name__ == "__main__":

    #
    #   call from main process to make sure works
    #
    info("starting")

    #
    #   call from child processes in pool
    #
    # start 4 worker processes
    pool = multiprocessing.Pool(processes=4)
    function_parameters = list()
    for a in range(200):
        function_parameters.append(("message #%3d" % a,))
    pool.map(test_logging, function_parameters)

    print(Logger.getCounts())



from CGATReportPlugins.Transformer import Transformer
from collections import OrderedDict as odict
from docutils.parsers.rst import directives
from CGATReport import DataTree, Utils
from CGATReport.DataTree import path2str
import itertools
import math
import scipy.stats


class TransformerHypergeometric(Transformer):

    '''Takes in lists of items (genes) and computes a hypergeometric p
    value and fold enrichment on the overlap between the lists.

    Takes bottom level of the data tree, which is expects to be a
    dictionary of lists, one of which should be named background. It
    will check if all the items in the other lists are also in the
    background list.

    If the dictionary contains more than 2 lists in addition to
    background, all pairwise overlaps are calculated.

    '''

    nlevels = -1

    def transform(self, data):

        # check if data is melted:
        if len(data.columns) != 1:
            raise ValueError(
                'transformer requires dataframe with'
                'a single column, got %s' % data.columns)
        column = data.columns[0]

        # iterate over lowest levels to build a dictionary of
        # sets
        genesets = {}
        nlevels = Utils.getDataFrameLevels(data)
        for key, group in data.groupby(level=range(nlevels)):
            genesets[path2str(key)] = set(group[column])

        keys = genesets.keys()

        background = None
        foreground = []
        for key in keys:
            if "background" in key:
                background = genesets[key]
            else:
                foreground.append(key)

        if len(keys) < 3 or background is None:
            raise ValueError(
                "Expected at least 3 lists, with one called background, "
                "instead got %i lists called %s" %
                (len(keys), ", ".join(keys)))

        missing = {
            y: [str(x) for x in genesets[y]
                if x not in background] for y in foreground}

        if any([len(missing[x]) > 0 for x in missing]):
            missing_items = "\n\t".join(
                ["%s:\t%s" % (x, ",".join(missing[x])) for x in missing])
            raise ValueError(
                "Found items in lists not in background. "
                "Missing items:\n\t %s" % missing_items)

        M = len(set(background))
        if len(keys) == 2:

            n = len(set(genesets[keys[1]]))
            N = len(set(genesets[keys[0]]))
            x = len(set(genesets[keys[0]]) & set(genesets[keys[1]]))

            p = scipy.stats.hypergeom.sf(x, M, n, N)

            fc = ((x + 0.0) / N) / ((n + 0.0) / M)

            values = [("Enrichment", fc),
                      ("P-value", p)]
        else:
            enrichments = []
            pvals = []
            As = []
            Bs = []
            for a, b in itertools.combinations(keys, 2):
                N = len(set(genesets[a]))
                n = len(set(genesets[b]))
                x = len(set(genesets[a]) & set(genesets[b]))

                p = scipy.stats.hypergeom.sf(x, M, n, N)

                fc = ((x + 0.0) / N) / ((n + 0.0) / M)

                As.append(a)
                Bs.append(b)
                pvals.append(p)
                enrichments.append(fc)

            values = [("ListA", As),
                      ("ListB", Bs),
                      ("Enrichment", enrichments),
                      ("P-value", pvals)]

        return DataTree.listAsDataFrame(values, values_are_rows=True)


class TransformerOddsRatio(Transformer):

    ''' Takes in lists of items (genes) and computes a hypergeometric p value and fold enrichment
    on the overlap between the lists.

    Takes bottom level of the data tree, which is expects to be a dictionary of lists, one of which
    should be named background. It will check if all the items in the other lists are also in
    the background list.

    If the dictionary contains more than 2 lists in addition to background, all pairwise overlaps
    are calculated. '''

    nlevels = 1

    def transform(self, data):

        keys = data.keys()
        if len(keys) < 3 or "background" not in data.keys():
            raise ValueError("Expected at least 3 lists, with one called background, instead got %i lists called %s" % (
                len(keys), ", ".join(keys)))

        keys = [x for x in keys if x != "background"]

        missing = {
            y: [str(x) for x in data[y] if x not in data["background"]] for y in keys}

        if any([len(missing[x]) > 0 for x in missing]):

            missing_items = "\n\t".join(
                ["%s:\t%s" % (x, ",".join(missing[x])) for x in missing])

            raise ValueError(
                "Found items in lists not in background. Missing items:\n\t %s" % missing_items)

        if len(keys) == 2:

            a = len(set(data[keys[0]]) & set(data[keys[1]]))
            b = len(set(data[keys[0]]) - set(data[keys[1]]))
            c = len(set(data[keys[1]]) - set(data[keys[0]]))
            d = len(
                set(data["background"]) - set(data[keys[0]]) - set(data[keys[1]]))

            table = [[a, b], [c, d]]

            OR, p = fisher_exact(table)

            sigma = math.sqrt((1.0 / a) + (1.0 / b) + (1.0 / c) + (1.0 / d))

            CI95_hi = OR * math.exp(1.96 * sigma)
            CI95_low = OR * math.exp(-1.96 * sigma)

            return odict([("OddRatio", OR),
                          ("CI95_hi", CI95_hi),
                          ("CI95_low", CI95_low),
                          ("P-value", p)])

        else:
            ORs = []
            pvals = []
            CIhis = []
            CIlos = []
            As = []
            Bs = []
            for x, z in itertools.combinations(keys, 2):

                a = len(set(data[x]) & set(data[z]))
                b = len(set(data[x]) - set(data[z]))
                c = len(set(data[z]) - set(data[x]))
                d = len(set(data["background"]) - set(data[x]) - set(data[z]))

                table = [[a, b], [c, d]]

                OR, p = fisher_exact(table)

                sigma = math.sqrt(
                    (1.0 / a) + (1.0 / b) + (1.0 / c) + (1.0 / d))

                CI95_hi = OR * math.exp(1.96 * sigma)
                CI95_low = OR * math.exp(-1.96 * sigma)

                As.append(x)
                Bs.append(z)
                pvals.append(p)
                ORs.append(OR)
                CIhis.append(CI95_hi)
                CIlos.append(CI95_low)

            return odict([("ListA", As),
                          ("ListB", Bs),
                          ("OddsRatio", ORs),
                          ("CI95_hi", CIhis),
                          ("CI95_low", CIlos),
                          ("P-value", pvals)])


class TransformerVenn(Transformer):
    '''Takes dictionarys of lists and transforms the data into the correct
    format for use with the venn renderer. If the dictionary contains
    a background entry and all other entries are subsets of the
    background, then it is removed, unless the keep-background options
    is specified.

    e.g.

    This:                           Becomes:
    a/x/[1,2,3]                     a/01/1
    a/y/[2,3,4]                     a/10/1
    a/background/[1,2,3,4,5]        a/11/2
    b/x/[5,6,7]                     b/01/1
    b/y/[6,7,8]                     b/10/1
                                    b/11/2

    Currently handles 2 or 3 way overlaps

    '''

    nlevels = -1

    options = Transformer.options + (
        ('keep-background', directives.flag),)

    background = False

    prune_dataframe = False

    def __init__(self, *args, **kwargs):
        Transformer.__init__(self, *args, **kwargs)

        self.background = "keep-background" in kwargs

    def transform(self, data):

        # check if data is melted:
        if len(data.columns) != 1:
            raise ValueError(
                'transformer requires dataframe with '
                'a single column, got %s' % data.columns)
        column = data.columns[0]
        # iterate over lowest levels to build a dictionary of
        # sets
        genesets = {}
        nlevels = Utils.getDataFrameLevels(data)
        for key, group in data.groupby(level=range(nlevels)):
            if "background" in key and not self.background:
                continue
            genesets[key] = set(group[column])
            
        values = []
        if len(genesets) == 2:
            a = set(genesets[genesets.keys()[0]])
            b = set(genesets[genesets.keys()[1]])

            values.append(("10", len(a - b)))
            values.append(("01", len(b - a)))
            values.append(("11", len(a & b)))
            values.append(("labels", map(path2str, genesets.keys())))
        elif len(genesets) == 3:
            a = set(genesets[genesets.keys()[0]])
            b = set(genesets[genesets.keys()[1]])
            c = set(genesets[genesets.keys()[2]])

            values.append(("100", len(a - b - c)))
            values.append(("010", len(b - a - c)))
            values.append(("001", len(c - a - b)))
            values.append(("110", len((a & b) - c)))
            values.append(("101", len((a & c) - b)))
            values.append(("011", len((b & c) - a)))
            values.append(("111", len((a & b) & c)))
            values.append(("labels", map(path2str, genesets.keys())))
        else:
            raise ValueError(
                "Can currently only cope with 2 or 3 way intersections")

        return DataTree.listAsDataFrame(values)


class TransformerMultiTest(Transformer):
    '''This transformer performs multiple testing correction.

    A multiple testing is applied to P-Values in a column called
    *P-value* (can be changed using the ``p-value`` option).

    The original data frame is returned with an added ``P-adjust``
    column. The default method for correction is BH, but other R style
    correction methods can be specified with ``adj-method`` option.

    P-values are adjusted overall.
    '''

    options = Transformer.options + (
        ('adj-method', directives.unchanged),
        ('p-value', directives.unchanged),)

    # group everything together
    nlevels = 0

    def __init__(self, *args, **kwargs):
        Transformer.__init__(self, *args, **kwargs)

        self.method = kwargs.get("adj-method", "BH")
        self.pval = kwargs.get("p-value", "P-value")

    def transform(self, data):

        from rpy2.robjects import r as R

        padj = R["p.adjust"](data[self.pval],
                             method=self.method)

        data["P-adjust"] = padj
        return data



import math
import numpy
import scipy
from functools import reduce

# See http://projects.scipy.org/scipy/ticket/1739
# scipy 0.11 for python3 broken, should be fixed for scipy 0.12
try:
    import scipy.stats
except ValueError:
    scipy.stats = None

try:
    from rpy2.robjects import r as R
    import rpy2.robjects.numpy2ri
except ImportError:
    R = None

from collections import OrderedDict as odict


def getSignificance(pvalue, thresholds=[0.05, 0.01, 0.001]):
    """return cartoon of significance of a p-Value."""
    n = 0
    for x in thresholds:
        if pvalue > x:
            return "*" * n
        n += 1
    return "*" * n


class Result(object):

    '''allow both member and dictionary access.'''
    slots = ("_data")

    def __init__(self):
        object.__setattr__(self, "_data", odict())

    def fromR(self, take, r_result):
        '''convert from an *r_result* dictionary using map *take*.

        *take* is a list of tuples mapping a field to the corresponding
        field in *r_result*.
        '''
        # deactivate numpy2ri conversion, interferes with .rx
        rpy2.robjects.numpy2ri.deactivate()
        for x, y in take:
            if y:
                try:
                    self._data[x] = r_result.rx(y)[0][0]
                except TypeError:
                    self._data[x] = "NA"
            else:
                self._data[x] = r_result.rx(x)[0][0]

        rpy2.robjects.numpy2ri.activate()

        return self

    def __len__(self):
        return self._data.__len__()

    def __getattr__(self, key):
        if not key.startswith("_"):
            try:
                return object.__getattribute__(self, "_data")[key]
            except KeyError:
                pass
        return getattr(self._data, key)

    def asDict(self):
        return self._data

    def keys(self):
        return list(self._data.keys())

    def values(self):
        return list(self._data.values())

    def __iter__(self):
        return self._data.__iter__()

    def __str__(self):
        return str(self._data)

    def __contains__(self, key):
        return key in self._data

    def __getitem__(self, key):
        return self._data[key]

    def __delitem__(self, key):
        del self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def __setattr__(self, key, value):
        if not key.startswith("_"):
            self._data[key] = value
        else:
            object.__setattr__(self, key, value)

    def __getstate__(self):
        # required for correct pickling/unpickling
        return object.__getattribute__(self, "_data")

    def __setstate__(self, d):
        # required for correct unpickling, otherwise
        # maximum recursion threshold will be reached
        object.__setattr__(self, "_data", d)

#################################################################
#################################################################
#################################################################
# Perform log likelihood test


class LogLikelihoodTest(Result):

    def __init__(self):
        pass


def doLogLikelihoodTest(complex_ll, complex_np,
                        simple_ll, simple_np,
                        significance_threshold=0.05):
    """perform log-likelihood test between model1 and model2.
    """

    assert complex_ll >= simple_ll, "log likelihood of complex model smaller than for simple model: %f > %f" % (
        complex_ll, simple_ll)

    chi = 2 * (complex_ll - simple_ll)
    df = complex_np - simple_np

    if df <= 0:
        raise ValueError("difference of degrees of freedom not larger than 0")

    p = scipy.stats.chisqprob(chi, df)

    l = LogLikelihoodTest()

    l.mComplexLogLikelihood = complex_ll
    l.mSimpleLogLikelihood = simple_ll
    l.mComplexNumParameters = complex_np
    l.mSimpleNumParameters = simple_np
    l.mSignificanceThreshold = significance_threshold
    l.mProbability = p
    l.mChiSquaredValue = chi
    l.mDegreesFreedom = df

    if p < significance_threshold:
        l.mPassed = True
    else:
        l.mPassed = False

    return l

#################################################################
#################################################################
#################################################################


class BinomialTest:

    def __init__(self):
        pass


def doBinomialTest(p, sample_size, observed, significance_threshold=0.05):
    """perform a binomial test.

    Given are p: the probability of the NULL hypothesis, the sample_size
    and the number of observed counts.
    """
    pass

#################################################################
#################################################################
#################################################################


class ChiSquaredTest:

    def __init__(self):
        pass


def doChiSquaredTest(matrix, significance_threshold=0.05):
    """perform chi-squared test on a matrix.
    """
    nrows, ncols = matrix.shape
    if nrows != 2 or ncols != 2:
        raise ValueError(
            "chi-square currently only implemented for 2x2 tables.")

    df = (nrows - 1) * (ncols - 1)

    row_sums = [matrix[x, :].sum() for x in range(nrows)]
    col_sums = [matrix[:, x].sum() for x in range(ncols)]
    sample_size = float(sum(row_sums))

    chi = 0.0
    for x in range(nrows):
        for y in range(ncols):
            expected = row_sums[x] * col_sums[y] / sample_size
            d = matrix[x, y] - expected
            chi += d * d / expected

    result = ChiSquaredTest()

    result.mProbability = scipy.stats.chisqprob(chi, df)
    result.mDegreesFreedom = df
    result.mChiSquaredValue = chi
    result.mPassed = result.mProbability < significance_threshold
    result.mSignificance = getSignificance(result.mProbability)
    result.mSampleSize = sample_size
    result.mPhi = math.sqrt(result.mChiSquaredValue / result.mSampleSize)
    return result


def doPearsonChiSquaredTest(p, sample_size, observed,
                            significance_threshold=0.05):
    """perform a pearson chi squared test.

    Given are p: the probability of the NULL hypothesis, the sample_size
    and the number of observed counts.

    For large sample sizes, this test is a continuous approximation to
    the binomial test.
    """
    e = float(p) * sample_size
    d = float(observed) - e
    chi = d * d / e
    df = 1

    result = ChiSquaredTest()

    result.mProbability = scipy.stats.chisqprob(chi, df)
    result.mDegreesFreedom = df
    result.mChiSquaredValue = chi
    result.mPassed = result.mProbability < significance_threshold
    result.mSignificance = getSignificance(result.mProbability)
    result.mSampleSize = sample_size
    result.mPhi = math.sqrt(result.mChiSquaredValue / result.mSampleSize)
    result.mObserved = observed
    result.mExpected = e
    return result

#################################################################
#################################################################
#################################################################
# Convenience functions and objects for statistical analysis


class Summary(Result):

    """a collection of distributional parameters. Available properties
    are:

    mean, median, min, max, samplestd, sum, counts
    """

    def __init__(self, values=None, format="%6.4f", mode="float"):
        Result.__init__(self)
        self._format = format
        self._mode = mode
        # note that this determintes the order of the fields at output
        self.counts, self.min, self.max, self.mean, self.median, self.samplestd, self.sum, self.q1, self.q3 = \
            (0, 0, 0, 0, 0, 0, 0, 0, 0)

        if values is not None:

            values = [x for x in values if x != None]

            if len(values) == 0:
                raise ValueError("no data for statistics")

            # convert
            self._nerrors = 0
            if type(values[0]) not in (int, float):
                n = []
                for x in values:
                    try:
                        n.append(float(x))
                    except ValueError:
                        self._nerrors += 1
            else:
                n = values

            if len(n) == 0:
                raise ValueError("no data for statistics")

            # use a non-sort algorithm later.
            n.sort()
            self.q1 = n[len(n) / 4]
            self.q3 = n[len(n) * 3 / 4]

            self.counts = len(n)
            self.min = min(n)
            self.max = max(n)
            self.mean = numpy.mean(n)
            self.median = numpy.median(n)
            self.samplestd = numpy.std(n)
            self.sum = reduce(lambda x, y: x + y, n)

    def getHeaders(self):
        """returns header of column separated values."""
        return ("nval", "min", "max", "mean", "median", "stddev", "sum", "q1", "q3")

    def getHeader(self):
        """returns header of column separated values."""
        return "\t".join(self.getHeaders())

    def __str__(self):
        """return string representation of data."""

        if self._mode == "int":
            format_vals = "%i"
            format_median = "%.1f"
        else:
            format_vals = self._format
            format_median = self._format

        return "\t".join(("%i" % self.counts,
                          format_vals % self.min,
                          format_vals % self.max,
                          self._format % self.mean,
                          format_median % self.median,
                          self._format % self.samplestd,
                          format_vals % self.sum,
                          format_vals % self.q1,
                          format_vals % self.q3,
                          ))


class FDRResult:

    def __init__(self):
        pass

    def plot(self, hardcopy=None):

        if hardcopy:
            R.png(hardcopy, width=1024, height=768, type="cairo")

        R.require('qvalue')

        # build a qobj
        R.assign("pval", self.mPValues)
        R.assign("pi0", self.mPi0)
        R.assign("qval", self.mQValues)
        R.assign("lambda", self.mLambda)
        R("""qobj <-list(pi0=pi0, qvalues=qval, pvalues=pval, lambda=lambda)""")
        R(""" class(qobj) <- "qvalue" """)

        R("""qplot(qobj)""")

        if hardcopy:
            R.dev_off()


def doFDR(pvalues,
          vlambda=numpy.arange(0, 0.95, 0.05),
          pi0_method="smoother",
          fdr_level=None,
          robust=False,
          smooth_df=3,
          smooth_log_pi0=False):
    """modeled after code taken from http://genomics.princeton.edu/storeylab/qvalue/linux.html.

    I did not like the error handling so I translated most to python.
    """

    if min(pvalues) < 0 or max(pvalues) > 1:
        raise ValueError("p-values out of range")

    if len(vlambda) > 1 and len(vlambda) < 4:
        raise ValueError(
            " If length of vlambda greater than 1, you need at least 4 values.")

    if len(vlambda) > 1 and (min(vlambda) < 0 or max(vlambda) >= 1):
        raise ValueError("vlambda must be within [0, 1).")

    m = len(pvalues)

    # these next few functions are the various ways to estimate pi0
    if len(vlambda) == 1:
        vlambda = vlambda[0]
        if vlambda < 0 or vlambda >= 1:
            raise ValueError("vlambda must be within [0, 1).")

        pi0 = numpy.mean([x >= vlambda for x in pvalues]) / (1.0 - vlambda)
        pi0 = min(pi0, 1.0)
        R.assign("pi0", pi0)
    else:
        pi0 = numpy.zeros(len(vlambda), numpy.float)

        for i in range(len(vlambda)):
            pi0[i] = numpy.mean([x >= vlambda[i]
                                 for x in pvalues]) / (1.0 - vlambda[i])

        R.assign("pi0", pi0)
        R.assign("vlambda", vlambda)

        if pi0_method == "smoother":
            if smooth_log_pi0:
                pi0 = math.log(pi0)

            R.assign("smooth_df", smooth_df)

            spi0 = R("""spi0 <- smooth.spline(vlambda,pi0, df = smooth_df)""")
            pi0 = R("""pi0 <- predict(spi0, x = max(vlambda))$y""")

            if smooth_log_pi0:
                pi0 = math.exp(pi0)

        elif pi0_method == "bootstrap":

            minpi0 = min(pi0)

            mse = numpy.zeros(len(vlambda), numpy.float)
            pi0_boot = numpy.zeros(len(vlambda), numpy.float)

            R.assign("pvalues", pvalues)
            pi0 = R("""
            m <- length(pvalues)
            minpi0 <- min(pi0)
            mse <- rep(0,length(vlambda))
            pi0_boot <- rep(0,length(vlambda))
            for(i in 1:100)
            {
                pvalues_boot <- sample(pvalues,size=m,replace=TRUE)
                for(i in 1:length(vlambda))
                {
                    pi0_boot[i] <- mean(pvalues_boot>vlambda[i])/(1-vlambda[i])
                }
                mse <- mse + (pi0_boot-minpi0)^2
            }
            pi0 <- min(pi0[mse==min(mse)])""")
        else:
            raise ValueError(
                "'pi0_method' must be one of 'smoother' or 'bootstrap'.")

        pi0 = min(pi0, 1.0)

    if pi0 <= 0:
        raise ValueError(
            "The estimated pi0 <= 0. Check that you have valid p-values or use another vlambda method.")

    if fdr_level != None and (fdr_level <= 0 or fdr_level > 1):
        raise ValueError("'fdr_level' must be within (0, 1].")

    # The estimated q-values calculated here
    #u = numpy.argsort(p)

    # change by Alan
    # ranking function which returns number of observations less than or equal
    R.assign("pvalues", pvalues)
    R.assign("robust", robust)
    qvalues = R("""u <- order(pvalues)
    qvalues.rank <- function(x)
{
      idx <- sort.list(x)

      fc <- factor(x)
      nl <- length(levels(fc))
      bin <- as.integer(fc)
      tbl <- tabulate(bin)
      cs <- cumsum(tbl)

      tbl <- rep(cs, tbl)
      tbl[idx] <- tbl

      return(tbl)
}

v <- qvalues.rank(pvalues)
m <- length(pvalues)

qvalues <- pi0 * m * pvalues / v
if(robust)
{
        qvalues <- pi0*m*pvalues/(v*(1-(1-pvalues)^m))
}
qvalues[u[m]] <- min(qvalues[u[m]],1)
for(i in (m-1):1)
{
   qvalues[u[i]] <- min(qvalues[u[i]],qvalues[u[i+1]],1)
}
qvalues
""")

    result = FDRResult()
    result.mQValues = qvalues

    if fdr_level != None:
        result.mPassed = [x <= fdr_level for x in result.mQValues]

    result.mPValues = pvalues
    result.mPi0 = pi0
    result.mLambda = vlambda

    return result

#################################################################
#################################################################
#################################################################


class CorrelationTest(Result):

    def __init__(self,
                 r_result=None,
                 s_result=None,
                 method=None,
                 nobservations=0):
        Result.__init__(self)

        self.pvalue = None
        self.method = method
        self.nobservations = 0

        if r_result:
            self.coefficient = r_result['estimate']['cor']
            self.pvalue = float(r_result['p.value'])
            self.nobservations = r_result['parameter']['df']
            self.method = r_result['method']
            self.alternative = r_result['alternative']
        elif s_result:
            self.coefficient = s_result[0]
            self.pvalue = s_result[1]
            self.nobservations = nobservations
            self.alternative = "two-sided"

        if self.pvalue != None:
            if self.pvalue > 0:
                self.logpvalue = math.log(self.pvalue)
            else:
                self.logpvalue = 0
            self.significance = getSignificance(self.pvalue)

    def __str__(self):
        return "\t".join((
            "%6.4f" % self.coefficient,
            "%e" % self.pvalue,
            self.significance,
            "%i" % self.nobservations,
            self.method,
            self.alternative))


def filterMasked(xvals, yvals, missing=("na", "Nan", None, ""), dtype = numpy.float):
    """convert xvals and yvals to numpy array skipping pairs with
    one or more missing values."""
    xmask = [i in missing for i in xvals]
    ymask = [i in missing for i in yvals]
    return (numpy.array([xvals[i] for i in range(len(xvals)) if not xmask[i]], dtype=dtype),
            numpy.array([yvals[i] for i in range(len(yvals)) if not ymask[i]], dtype=dtype))


def filterNone(args, missing=("na", "Nan", None, "", 'None', 'none'), dtype = numpy.float):
    '''convert arrays in 'args' to numpy arrays of 'dtype', skipping where any of
    the columns have a value of missing.

    >>> Stats.filterNone(((1,2,3), (4,5,6)))
    [array([ 1.,  2.,  3.]), array([ 4.,  5.,  6.])]
    >>> Stats.filterNone(((1,2,3), (4,None,6)))
    [array([ 1.,  3.]), array([ 4.,  6.])]
    >>> Stats.filterNone(((None,2,3), (4,None,6)))
    [array([ 3.]), array([ 6.])]
    '''
    mi = min([len(x) for x in args])
    ma = max([len(x) for x in args])
    assert mi == ma, "arrays have unequal length to start with: min=%i, max=%i." % (
        mi, ma)

    mask = [sum([z in missing for z in x]) for x in zip(*args)]

    return [numpy.array([x[i] for i in range(len(x)) if not mask[i]], dtype=dtype) for x in args]


def filterMissing(args,
                  missing=("na", "Nan", None, "", 'None', 'none'),
                  dtype = numpy.float):
    '''remove rows in args where at least one of the columns have a
       missing value.'''

    mi = min([len(x) for x in args])
    ma = max([len(x) for x in args])
    assert mi == ma, "arrays have unequal length to start with: min=%i, max=%i." % (
        mi, ma)

    keep = numpy.array([True] * ma)
    for values in args:
        keep &= values.notnull()

    return [x[keep] for x in args]


def doCorrelationTest(xvals, yvals, method="pearson"):
    """compute correlation between x and y.

    Raises a value-error if there are not enough observations.
    """

    if scipy.stats is None:
        raise ImportError("scipy.stats not available")

    if len(xvals) <= 1 or len(yvals) <= 1:
        raise ValueError("can not compute correlation with no data")
    if len(xvals) != len(yvals):
        raise ValueError("data vectors have unequal length")

    x, y = filterMasked(xvals, yvals)

    if method == "pearson":
        s_result = scipy.stats.pearsonr(x, y)
    elif method == "spearman":
        s_result = scipy.stats.spearmanr(x, y)
    else:
        raise ValueError("unknown method %s" % (method))

    result = CorrelationTest(s_result=s_result,
                             method=method,
                             nobservations=len(x))

    return result.asDict()


###################################################################
###################################################################
###################################################################
# compute ROC curves from sorted values
###################################################################
def computeROC(values):
    '''return a roc curve for *values*.
    Values is a sorted list of (value, bool) pairs.

    '''
    roc = []

    npositives = len([x for x in values if x[1]])
    if npositives == 0:
        raise ValueError("no positives among values")

    ntotal = len(values)

    last_value, last_fpr = None, None
    tp, fp = 0, 0
    tn, fn = ntotal - npositives, npositives

    for value, is_positive in values:
        if is_positive:
            tp += 1
            fn -= 1
        else:
            fp += 1
            tn -= 1

        if last_value != value:

            try:
                tpr = float(tp) / (tp + fn)
            except ZeroDivisionError:
                tpr = 0

            try:
                fpr = float(fp) / (fp + tn)
            except ZeroDivisionError:
                fpr = 0

            if last_fpr != fpr:
                roc.append((fpr, tpr))
                last_fpr = fpr

        last_values = value

    return roc

###################################################################
###################################################################
###################################################################
##
###################################################################


def getAreaUnderCurve(xvalues, yvalues):
    '''compute area under curve from a set of discrete x,y coordinates
    using trapezoids.

    This is only as accurate as the density of points.
    '''

    assert len(xvalues) == len(yvalues)
    last_x, last_y = xvalues[0], yvalues[0]
    auc = 0
    for x, y in zip(xvalues, yvalues)[1:]:
        dx = x - last_x
        assert not dx < 0, "x not increasing: %f >= %f" % (last_x, x)
        dy = abs(last_y - y)
        my = min(last_y, y)
        # rectangle plus triangle
        auc += dx * my + dx * dy / 2
        last_x, last_y = x, y

    return auc

###################################################################
###################################################################
###################################################################
##
###################################################################


def getSensitivityRecall(values):
    '''return sensitivity/selectivity.

    Values is a sorted list of (value, bool) pairs.
    '''

    npositives = 0.0
    npredicted = 0.0
    l = None
    result = []
    total = float(len(values))
    for value, is_positive in values:
        npredicted += 1.0
        if is_positive > 0:
            npositives += 1.0
        if value != l:
            result.append((value, npositives / npredicted, npredicted / total))
        l = value
    if l:
        result.append((l, npositives / npredicted, npredicted / total))

    return result.asDict()

###################################################################
###################################################################
###################################################################
##
###################################################################


def doMannWhitneyUTest(xvals, yvals):
    '''apply the Mann-Whitney U test to test for the difference of medians.'''
    if len(xvals) == 0 or len(yvals) == 0:
        result = Result()
    else:
        if R:
            r_result = R['wilcox.test'](xvals, yvals, paired=False)
            result = Result().fromR(
                (("pvalue", 'p.value'),
                 ('alternative', None),
                 ('method', None)),
                r_result)
        else:
            raise ValueError("rpy2 not available")

    result.xobservations = len(xvals)
    result.yobservations = len(yvals)

    return result.asDict()


def buildMatrixFromEdges(edges,
                         in_map_token2row={},
                         in_map_token2col={},
                         is_square=True,
                         is_symmetric=False,
                         missing_value=0,
                         diagonal_value=0,
                         dtype=numpy.int):
    """build a matrix from an edge-list representation.

    For example, the following list of tuples::

       [('A', 'B', 1),
        ('A', 'C', 2),
        ('B', 'C', 3)]

    will be converted to the following matrix::

         A B C
       A   1 2
       B     3
       C

    If *is_symmetric* is set to True, the matrix is assumed to be
    symmetric and missing values will automatically be filled::

         A B C
       A   1 2
       B 1   3
       C 2 3

    If edge list may contain four elements, in which case the
    fourth element is expected to be the value of the lower
    diagonal in a symmetric matrix::

       [('A', 'B', 1, 4),
        ('A', 'C', 2, 5),
        ('B', 'C', 3, 6)]

    will yield::

         A B C
       A   1 2
       B 4   3
       C 5 6

    If *is_square* the matrix will be squared.

    returns a numpy matrix and lists of row and column names.
    """

    if in_map_token2row:
        map_token2row = in_map_token2row
    else:
        map_token2row = {}

    if in_map_token2col:
        map_token2col = in_map_token2col
    else:
        map_token2col = {}

    has_row_names = len(map_token2row) > 0
    has_col_names = len(map_token2col) > 0

    # if either row/column names are not given:
    if not map_token2row or not map_token2col:

        row_tokens = sorted(list(set([x[0] for x in edges])))
        col_tokens = sorted(list(set([x[1] for x in edges])))

        if not has_row_names:
            for row_token in row_tokens:
                if row_token not in map_token2row:
                    map_token2row[row_token] = len(map_token2row)
        if not has_col_names:
            for col_token in col_tokens:
                if col_token not in map_token2col:
                    map_token2col[col_token] = len(map_token2col)

        # for square matrices merge row and column labels
        if is_square:
            for col_token in map_token2col.keys():
                if col_token not in map_token2row:
                    map_token2row[col_token] = len(map_token2row)
            map_token2col = map_token2row

    matrix = numpy.matrix([missing_value] *
                          len(map_token2row) * len(map_token2col),
                          dtype=dtype)
    matrix.shape = (len(map_token2row), len(map_token2col))

    if is_square:
        for i in range(len(map_token2col)):
            matrix[i, i] = diagonal_value

    if len(edges[0]) == 3:
        if is_symmetric:
            for row, col, value in edges:
                if value is not None:
                    matrix[map_token2row[row], map_token2col[col]] = \
                        matrix[map_token2row[col], map_token2col[row]] = value
        else:
            for row, col, value in edges:
                if value is not None:
                    matrix[map_token2row[row], map_token2col[col]] = value
    elif len(edges[0]) == 4:
        for row, col, value1, value2 in edges:
            if value1 is not None:
                matrix[map_token2row[row], map_token2col[col]] = value1
            if value2 is not None:
                matrix[map_token2row[col], map_token2col[row]] = value2
    else:
        raise ValueError(
            "unexpected number of elements in list, expected 3 or 4, "
            "got %i" % (len(edges[0])))

    col_tokens = map_token2col.items()
    col_tokens.sort(lambda x, y: cmp(x[1], y[1]))
    row_tokens = map_token2row.items()
    row_tokens.sort(lambda x, y: cmp(x[1], y[1]))

    return matrix, [x[0] for x in row_tokens], [x[0] for x in col_tokens]



from CGATReport.Tracker import Tracker


class MyDataOneTrack(Tracker):

    """My tracked data - one track."""

    def getTracks(self):
        return ["all", ]

    def __call__(self, track):
        return dict((("header1", 10), ("header2", 20)),)


class MyDataTwoTracks(Tracker):

    """My tracked data - two tracks."""

    def getTracks(self):
        return ["track1", "track2"]

    def __call__(self, track):
        if track == "track1":
            return dict((("header1", 10), ("header2", 20)),)
        elif track == "track2":
            return dict((("header1", 20), ("header2", 10)),)



import re
import traceback
import itertools
import pandas

from CGATReport.ResultBlock import ResultBlocks
from CGATReport import DataTree
from CGATReport import Component
from CGATReport import Utils
from CGATReport import Cache

# move User renderer to CGATReport main distribution
from CGATReportPlugins import Renderer

VERBOSE = True
# maximimum number of levels in data tree
MAX_PATH_NESTING = 5

from collections import OrderedDict as odict

# heap memory debugging, search for 'heap' in this code
# from guppy import hpy; HP=hpy()


class Dispatcher(Component.Component):

    """Dispatch the directives in the ``:report:`` directive
    to a:class:`Tracker`, class:`Transformer` and:class:`Renderer`.

    The dispatcher has three passes:

    1. Collect data from a:class:`Tracker`

    2. Transform data from a:class:`Transformer`

    3. Render data with a:class:`Renderer`

    This class adds the following options to the:term:`render` directive.

    :term:`groupby`: group data by:term:`track` or:term:`slice`.

    """

    def __init__(self, tracker, renderer,
                 transformers=None):
        '''render images using an instance of a:class:`Tracker.Tracker`
        to obtain data, an optional:class:`Transformer` to transform
        the data and a:class:`Renderer.Renderer` to render the output.
        '''
        Component.Component.__init__(self)

        self.debug("starting dispatcher '%s': tracker='%s', renderer='%s', "
                   "transformer:='%s'" %
                   (str(self), str(tracker), str(renderer), str(transformers)))

        self.tracker = tracker
        # add reference to self for access to tracks
        self.tracker.dispatcher = self
        self.renderer = renderer
        self.transformers = transformers

        # set to true if index will be later set by tracker
        self.indexFromTracker = False

        try:
            self.debug("cache of tracker: %s: %s" % (self.tracker,
                                                     str(tracker.cache)))

            if tracker.cache:
                self.cache = Cache.Cache(Cache.tracker2key(tracker))
            else:
                self.cache = {}
                self.nocache = True
        except AttributeError:
            self.cache = Cache.Cache(Cache.tracker2key(tracker))

        self.tree = None
        self.data = None

        # Level at which to group the results of Renderers
        # None is no grouping
        # 0: group on first level ('groupby=track')
        # 1: group on second level ('groupby=slice')
        # n: group on n-th level
        self.group_level = 0

    def __del__(self):
        pass

    def getCaption(self):
        """return a caption string."""
        return self.__doc__

    def parseArguments(self, *args, **kwargs):
        '''argument parsing.'''

        self.groupby = kwargs.get("groupby", "default")
        try:
            self.groupby = int(self.groupby)
        except ValueError:
            pass

        self.nocache = "nocache" in kwargs

        try:
            self.mInputTracks = [x.strip()
                                 for x in kwargs["tracks"].split(",")]
        except KeyError:
            self.mInputTracks = None

        try:
            self.mInputSlices = [x.strip()
                                 for x in kwargs["slices"].split(",")]
        except KeyError:
            self.mInputSlices = None

        try:
            self.restrict_paths = [x.strip()
                                   for x in kwargs["restrict"].split(",")]
        except KeyError:
            self.restrict_paths = None

        try:
            self.exclude_paths = [x.strip()
                                  for x in kwargs["exclude"].split(",")]
        except KeyError:
            self.exclude_paths = None

        try:
            self.mColumns = [x.strip() for x in kwargs["columns"].split(",")]
        except KeyError:
            self.mColumns = None

        self.tracker_options = kwargs.get("tracker", None)

    def getData(self, path):
        """get data for track and slice. Save data in persistent cache for further use.

        For functions, path should be an empty tuple.
        """

        if path:
            key = DataTree.path2str(path)
        else:
            key = "all"

        result, fromcache = None, False
        # trackers with options are not cached
        if not self.nocache and not self.tracker_options:
            try:
                result = self.cache[key]
                fromcache = True
            except KeyError:
                pass
            except RuntimeError as msg:
                raise RuntimeError(
                    "error when accessing key %s from cache: %s "
                    "- potential problem with unpickable object?" % (key, msg))

        kwargs = {}
        if self.tracker_options:
            kwargs['options'] = self.tracker_options

        if result is None:
            try:
                result = self.tracker(*path, **kwargs)
            except Exception as msg:
                self.warn("exception for tracker '%s', path '%s': msg=%s" %
                          (str(self.tracker),
                           DataTree.path2str(
                               path),
                           msg))
                if VERBOSE:
                    self.warn(traceback.format_exc())
                raise

        # store in cache
        if not self.nocache and not fromcache:
            # exception - do not store data frames
            # test with None fails for some reason
            self.cache[key] = result

        return result

    def getDataPaths(self, obj):
        '''determine if obj is a function and return
        data paths from a tracker.

        If obj is a function, returns True and an empty list.
        If obj is an object, returns False and a list datapaths.
        '''
        data_paths = []

        if hasattr(obj, 'paths'):
            data_paths = getattr(obj, 'paths')
        elif hasattr(obj, 'getPaths'):
            data_paths = obj.getPaths()

        if not data_paths:
            if hasattr(obj, 'setIndex'):
                self.indexFromTracker = True
                return True, []
            elif hasattr(obj, 'tracks'):
                tracks = getattr(obj, 'tracks')
            elif hasattr(obj, 'getTracks'):
                tracks = obj.getTracks()
            else:
                # is a function
                return True, []

            # if tracks specified and no tracks found - return
            # no data paths.
            if tracks is None or len(tracks) == 0:
                return False, []

            data_paths = [tracks]

            # get slices
            if hasattr(obj, 'slices'):
                data_paths.append(getattr(obj, 'slices'))
            elif hasattr(obj, 'getSlices'):
                data_paths.append(obj.getSlices())

        # sanity check on data_paths.
        # 1. Replace strings with one-element tuples
        # 2. Remove empty levels in the paths
        # 3. Replace sets and other non-containers with lists
        to_remove = []
        for x, y in enumerate(data_paths):
            if y is None or len(y) == 0:
                to_remove.append(x)
                continue
            if isinstance(y, str):
                data_paths[x] = [y, ]
            elif type(y) not in Utils.ContainerTypes:
                data_paths[x] = list(y)

        for x in to_remove[::-1]:
            del data_paths[x]

        return False, data_paths

    def filterDataPaths(self, datapaths):
        '''filter *datapaths*.

        returns the filtered data paths.
        '''

        if not datapaths:
            return

        def _filter(all_entries, input_list):
            # need to preserve type of all_entries
            result = []
            search_entries = list(map(str, all_entries))
            for s in input_list:
                if s in search_entries:
                    # collect exact matches
                    result.append(all_entries[search_entries.index(s)])
                elif s.startswith("r(") and s.endswith(")"):
                    # collect pattern matches:
                    # remove r()
                    s = s[2:-1]
                    # remove flanking quotation marks
                    if s[0] in ('"', "'") and s[-1] in ('"', "'"):
                        s = s[1:-1]
                    rx = re.compile(s)
                    result.extend(
                        [all_entries[y] for y, x in
                         enumerate(search_entries) if rx.search(str(x))])
            return result

        if self.mInputTracks:
            datapaths[0] = _filter(datapaths[0], self.mInputTracks)

        if self.mInputSlices:
            if len(datapaths) >= 2:
                datapaths[1] = _filter(datapaths[1], self.mInputSlices)

        return datapaths

    def collect(self):
        '''collect all data.

        Data is stored in a multi-level dictionary (DataTree)
        '''

        self.tree = odict()

        self.debug("%s: collecting data paths." % (self.tracker))
        is_function, datapaths = self.getDataPaths(self.tracker)
        self.debug("%s: collected data paths." % (self.tracker))

        # if function, no datapaths
        if is_function:
            d = self.getData(())

            # save in data tree as leaf
            DataTree.setLeaf(self.tree, ("all",), d)

            self.debug("%s: collecting data finished for function." %
                       (self.tracker))
            return

        # if no tracks, error
        if len(datapaths) == 0 or len(datapaths[0]) == 0:
            self.warn("%s: no tracks found - no output" % self.tracker)
            return

        # filter data paths
        self.debug("%s: filtering data paths: %s" %
                   (self.tracker, datapaths))
        datapaths = self.filterDataPaths(datapaths)
        self.debug("%s: filtered data paths: %s" %
                   (self.tracker, datapaths))

        # if no tracks, error
        if len(datapaths) == 0 or len(datapaths[0]) == 0:
            self.warn(
                "%s: no tracks remain after filtering "
                "- no output" % self.tracker)
            return

        self.debug("%s: building all_paths" % (self.tracker))
        if len(datapaths) > MAX_PATH_NESTING:
            self.warn("%s: number of nesting in data paths too large: %i" % (
                self.tracker, len(datapaths)))
            raise ValueError(
                "%s: number of nesting in data paths too large: %i" % (
                    self.tracker, len(datapaths)))

        all_paths = list(itertools.product(*datapaths))
        self.debug(
            "%s: collecting data started for %i data paths" % (
                self.tracker,
                len(all_paths)))

        self.tree = odict()
        for path in all_paths:

            d = self.getData(path)

            # ignore empty data sets
            if d is None:
                continue

            # save in data tree as leaf
            DataTree.setLeaf(self.tree, path, d)

        self.debug(
            "%s: collecting data finished for %i data paths" % (
                self.tracker,
                len(all_paths)))
        return self.tree

    def _match(self, label, paths):
        '''return True if any of paths match to label.'''

        for s in paths:
            if label == s:
                return True
            elif s.startswith("r(") and s.endswith(")"):
                    # collect pattern matches:
                    # remove r()
                    s = s[2:-1]
                    # remove flanking quotation marks
                    if s[0] in ('"', "'") and s[-1] in ('"', "'"):
                        s = s[1:-1]
                    rx = re.compile(s)
                    if not Utils.isString(label):
                        continue
                    if rx.search(label):
                        return True
        return False

    def filterPaths(self, path_patterns, mode="restrict"):
        '''restrict or exclude data paths.

        Only those data paths and columns matching the restrict term
        are accepted.

        Columns are not removed.
        '''
        if not path_patterns:
            return

        # rows first

        # select rows to keep (matching any of the patterns in any
        # of the levels of the hierarchical index)
        is_hierarchical = isinstance(self.data.index,
                                     pandas.core.index.MultiIndex)
        if is_hierarchical:
            keep = [any([self._match(x, path_patterns) for x in labels])
                    for labels in self.data.index]
        else:
            keep = [self._match(x, path_patterns) for x in self.data.index]

        if mode == "exclude":
            keep = [not x for x in keep]

        self.data = self.data[keep]

        # Selecting columns does not work together with row restrict
        # needs a separate option.
        # keep = [x for x in self.data.columns if _match(x)]
        # self.data = self.data[keep]

    def transform(self):
        '''call data transformers and group tree
        '''
        for transformer in self.transformers:
            self.debug("profile: started: transformer: %s" % (transformer))
            self.debug("%s: applying %s" % (self.renderer, transformer))
            try:
                self.data = transformer(self.data)
            finally:
                self.debug("profile: finished: transformer: %s" %
                           (transformer))

        return self.data

    def group(self):
        '''rearrange dataframe for desired grouping.

        Through grouping the dataframe is rearranged such that the
        level at which data will be grouped will be the first level in
        hierarchical index.

        If grouping by "track" is set, additional level will be added
        to ensure that grouping will happen.

        If grouping by "slice" is set, the first two levels will
        be swopped.

        The group level indicates at which level in the nested dictionary
        the data will be grouped with 0 indicating that everything will
        grouped together.
        '''

        nlevels = Utils.getDataFrameLevels(self.data)
        try:
            default_level = self.renderer.group_level
        except AttributeError:
            # User rendere that is pure functions does not
            # have a group_level attribute
            default_level = 0

        groupby = self.groupby

        if str(default_level).startswith("force"):
            groupby = default_level[len("force-"):]
        elif groupby == "default":
            groupby = default_level

        if groupby == "none":
            self.group_level = nlevels - 1

        elif groupby == "track":
            self.group_level = 0

        elif groupby == "slice":
            # rearrange first two levels in data tree
            if nlevels > 1:
                self.data = self.data.reorder_levels(
                    [1, 0] + range(2, nlevels))
            self.group_level = 0

        elif groupby == "all":
            # group everything together
            self.group_level = -1

        elif isinstance(groupby, int):
            # get group level from Renderer
            if groupby < 0:
                # negative levels - subtract from lowest
                # level
                g = nlevels + groupby - 1
            else:
                g = groupby
            self.group_level = g
        else:
            self.group_level = 0

        self.debug("grouping: nlevel=%i, groupby=%s, default=%s, group=%i" %
                   (nlevels, groupby, str(default_level), self.group_level))

        return self.data

    def prune(self):
        '''prune data frame.

        Remove levels from the data tree that are
        superfluous, i.e. levels that contain only a single label.

        This method ignores some labels with reserved key-words
        such as ``text``, ``rst``, ``xls``

        '''

        dataframe = self.data
        self.pruned = []

        if isinstance(dataframe.index,
                      pandas.core.index.MultiIndex):
            todrop = []
            for x, level in enumerate(dataframe.index.levels):

                if len(level) == 1 and level[0] not in Utils.TrackerKeywords:
                    todrop.append(x)

            names = dataframe.index.names
            pruned = [(x, names[x]) for x in todrop]

            dataframe.reset_index(
                todrop,
                drop=True,
                inplace=True)

            for level, label in pruned:
                self.debug("pruned level %i from data tree: label='%s'" %
                           (level, label))

            # save for conversion
            self.pruned = pruned

    def render(self):
        '''supply the:class:`Renderer.Renderer` with the data to render.

        The data supplied will depend on the ``groupby`` option.

        returns a ResultBlocks data structure.
        '''
        self.debug("%s: rendering data started for %i items" %
                   (self,
                    len(self.data)))

        # initiate output structure
        results = ResultBlocks(title="")

        dataframe = self.data

        # dataframe.write_csv("test.csv")

        if dataframe is None:
            self.warn("%s: no data after conversion" % self)
            raise ValueError("no data for renderer")

        # special patch: set column names to pruned levels
        # if there are no column names
        if len(dataframe.columns) == len(self.pruned):
            if list(dataframe.columns) == list(range(len(dataframe.columns))):
                dataframe.columns = [x[1] for x in self.pruned]

        nlevels = Utils.getDataFrameLevels(dataframe)

        self.debug("%s: rendering data started. "
                   "levels=%i, group_level=%s" %
                   (self, nlevels,
                    str(self.group_level)))

        if self.group_level < 0:
            # no grouping for renderers that will accept
            # a dataframe with any level of indices and no explicit
            # grouping has been asked for.
            results.append(self.renderer(dataframe, path=()))
        else:
            level = Utils.getGroupLevels(
                dataframe,
                max_level=self.group_level+1)

            self.debug("%s: grouping by levels: %s" %
                       (self, str(level)))

            for key, work in dataframe.groupby(level=level):

                try:
                    results.append(self.renderer(work,
                                                 path=key))
                except:
                    self.error("%s: exception in rendering" % self)
                    results.append(
                        ResultBlocks(Utils.buildException("rendering")))

        if len(results) == 0:
            self.warn("renderer returned no data.")
            raise ValueError("renderer returned no data.")

        self.debug("%s: rendering data finished with %i blocks" %
                   (self.tracker, len(results)))

        return results

    def __call__(self, *args, **kwargs):

        try:
            self.parseArguments(*args, **kwargs)
        except:
            self.error("%s: exception in parsing" % self)
            return ResultBlocks(ResultBlocks(Utils.buildException("parsing")))

        # collect no data if tracker is the empty tracker
        # and go straight to rendering
        try:
            if self.tracker.getTracks() == ["empty"]:
                # is instance does not work because of module mapping
                # type(Tracker.Empty) == CGATReport.Tracker.Empty
                # type(self.tracker) == Tracker.Empty
                # if isinstance(self.tracker, Tracker.Empty):
                result = self.renderer()
                return ResultBlocks(result)
        except AttributeError:
            # for function trackers
            pass

        self.debug("profile: started: tracker: %s" % (self.tracker))

        # collecting data
        try:
            self.collect()
        except:
            self.error("%s: exception in collection" % self)
            return ResultBlocks(ResultBlocks(
                Utils.buildException("collection")))
        finally:
            self.debug("profile: finished: tracker: %s" % (self.tracker))

        if self.tree is None or len(self.tree) == 0:
            self.info("%s: no data - processing complete" % self.tracker)
            return None

        data_paths = DataTree.getPaths(self.tree)
        self.debug("%s: after collection: %i data_paths: %s" %
                   (self, len(data_paths), str(data_paths)))

        # special Renderers - do not process data further but render
        # directly. Note that no transformations will be applied.
        if isinstance(self.renderer, Renderer.User):
            results = ResultBlocks(title="main")
            results.append(self.renderer(self.tree))
            return results
        elif isinstance(self.renderer, Renderer.Debug):
            results = ResultBlocks(title="main")
            results.append(self.renderer(self.tree))
            return results

        # merge all data to hierarchical indexed dataframe
        self.data = DataTree.asDataFrame(self.tree)

        self.debug("dataframe memory usage: total=%i,data=%i,index=%i,col=%i" %
                   (self.data.values.nbytes +
                    self.data.index.nbytes +
                    self.data.columns.nbytes,
                    self.data.values.nbytes,
                    self.data.index.nbytes,
                    self.data.columns.nbytes))

        # if tracks are set by tracker, call tracker with dataframe
        if self.indexFromTracker:
            self.tracker.setIndex(self.data)

        # transform data
        try:
            self.transform()
        except:
            self.error("%s: exception in transformation" % self)
            return ResultBlocks(ResultBlocks(
                Utils.buildException("transformation")))

        # data_paths = DataTree.getPaths(self.data)
        # self.debug("%s: after transformation: %i data_paths: %s" %
        #           (self, len(data_paths), str(data_paths)))
        # restrict
        try:
            self.filterPaths(self.restrict_paths, mode="restrict")
        except:
            self.error("%s: exception in restrict" % self)
            return ResultBlocks(ResultBlocks(
                Utils.buildException("restrict")))

        # data_paths = DataTree.getPaths(self.data)
        # self.debug("%s: after restrict: %i data_paths: %s" %
        #          (self, len(data_paths), str(data_paths)))
        # exclude
        try:
            self.filterPaths(self.exclude_paths, mode="exclude")
        except:
            self.error("%s: exception in exclude" % self)
            return ResultBlocks(ResultBlocks(Utils.buildException("exclude")))

        # data_paths = DataTree.getPaths(self.data)
        # self.debug("%s: after exclude: %i data_paths: %s" %
        #          (self, len(data_paths), str(data_paths)))

        # No pruning - maybe enable later as a user option
        self.pruned = []
        # try:
        #     self.prune()
        # except:
        #     self.error("%s: exception in pruning" % self)
        #     return ResultBlocks(ResultBlocks(Utils.buildException("pruning")))

        # data_paths = DataTree.getPaths(self.data)
        # self.debug("%s: after pruning: %i data_paths: %s" %
        #           (self, len(data_paths), str(data_paths)))
        try:
            self.group()
        except:
            self.error("%s: exception in grouping" % self)
            return ResultBlocks(ResultBlocks(Utils.buildException("grouping")))

        # data_paths = DataTree.getPaths(self.data)
        # self.debug("%s: after grouping: %i data_paths: %s" %
        #           (self, len(data_paths), str(data_paths)))
        if self.renderer is not None:
            self.debug("profile: started: renderer: %s" % (self.renderer))

            try:
                result = self.render()
            except:
                self.error("%s: exception in rendering" % self)
                return ResultBlocks(ResultBlocks(
                    Utils.buildException("rendering")))
            finally:
                self.debug("profile: finished: renderer: %s" % (self.renderer))
        else:
            result = ResultBlocks(title="")

        return result

    def getTracks(self):
        '''return tracks used by the dispatcher.

        used for re-constructing call to cache.
        '''
        return self.tracks

    def getSlices(self):
        '''return slices used by the dispatcher.

        used for re-constructing call to cache.
        '''
        return self.slices
        
    def getDataTree(self):
        '''return data tree.'''
        return self.tree

    def getDataFrame(self):
        '''return data frame.'''
        return self.data



#! /bin/env python

"""cgatreport-test
=================

:file:`cgatreport-test` permits testing:class:`Trackers` and
:class:`Renderers` before using them in documents. The full list of
command line options is available using the:option:`-h/--help` command line
options.

The options are:

**-t/--tracker** tracker
:term:`tracker` to use.

**-r/--renderer** renderer
:term:`renderer` to use.

**-f/--force**
   force update of a:class:`Tracker`. Removes all data from cache.

**-m/--transformer** transformer
   :class:`Transformer` to use. Several transformers can be applied
   via multiple **-m** options.

**-a/--tracks** tracks
   Tracks to display as a comma-separated list.

**-s/--slices** slices
   Slices to display as a comma-separated list.

**-o/--option** option
   Options for the renderer/transformer. These correspond to options
   within restructured text directives, but supplied as key=value
   pairs (without spaces).  For example: ``:width: 300`` will become
   ``-o width=300``. Several **-o** options can be supplied on the
   command line.

**--no-print**
   Do not print an rst text template corresponding to the displayed plots.

**--no-show**
   Do not show plot. Use this to just display the tracks/slices that
   will be generated.

**-w/--path** path
   Path with trackers. By default,:term:`trackers` are searched in the
   directory:file`trackers` within the current directory.

**-i/--interactive**
   Start python interpreter.

If no command line arguments are given all:term:`trackers` are build
in parallel.

Usage
-----

There are three main usages of:command:`cgatreport-test`:

Fine-tuning plots
+++++++++++++++++

Given a:class:`Tracker` and:class:`Renderer`, cgatreport-test
will call the:class:`Tracker` and supply it to the:class:`Renderer`::

   cgatreport-test tracker renderer

Use this method to fine-tune plots. Additional options can be supplied
using the ``-o`` command line option. The script will output a template
restructured text snippet that can be directly inserted into a document.

Rendering a document
++++++++++++++++++++

With the ``-p/--page`` option, ``cgatreport-test`` will create the
restructured text document as it is supplied to sphinx::

   cgatreport-test --page=example.rst

This functionality is useful for debugging.

Testing trackers
++++++++++++++++

Running cgatreport-test without options::

   cgatreport-test

will collect all:class:`Trackers` and will execute them.
Use this method to see if all:class:`Trackers` can access
their data sources.

"""
import sys
import os
import re
import glob
import optparse
import tempfile
import logging as L
import code

import matplotlib.pyplot as plt
from matplotlib import _pylab_helpers

from CGATReport import Utils

import CGATReport.clean
from CGATReport.Dispatcher import Dispatcher
from CGATReport.Component import getOptionMap

# import conf.py
if os.path.exists("conf.py"):
    try:
        exec(compile(open("conf.py").read(), "conf.py", 'exec'))
    except ValueError:
        pass

# set default directory where trackers can be found
TRACKERDIR = "trackers"
if "docsdir" in locals():
    TRACKERDIR = os.path.join(docsdir, "trackers")

RST_TEMPLATE = """.. report:: %(tracker)s
   :render: %(renderer)s
   %(options)s

   %(caption)s
"""


def getTrackers(fullpath):
    """retrieve a tracker and its associated code.

    The tracker is not instantiated.

    returns a tuple (code, tracker, module, flag).

    The flag indicates whether that tracker is derived from the
    tracker base class. If False, that tracker will not be listed
    as an available tracker, though it might be specified at the
    command line.
    """

    name, cls = os.path.splitext(fullpath)
    # remove leading '.'
    cls = cls[1:]

    module_name = os.path.basename(name)
    module, pathname = Utils.getModule(name)
    trackers = []

    for name in dir(module):
        obj = getattr(module, name)

        try:
            if Utils.isClass(obj):
                trackers.append((name, obj, module_name, True))
            else:
                trackers.append((name, obj, module_name, False))
        except ValueError:
            pass

    return trackers


def writeRST(outfile, options, kwargs,
             renderer_options, transformer_options,
             display_options,
             modulename, name):
    '''write RST snippet to outfile to be include in a cgatreport document
    '''

    options_rst = []

    for key, val in list(kwargs.items()) +\
            list(renderer_options.items()) +\
            list(transformer_options.items()) +\
            list(display_options.items()):

        if val is None:
            options_rst.append(":%s:" % key)
        else:
            options_rst.append(":%s: %s" % (key, val))

    params = {"tracker": "%s.%s" % (modulename, name),
              "renderer": options.renderer,
              "options": ("\n   ").join(options_rst),
              "caption": options.caption}
    if options.transformers:
        params["options"] = ":transform: %s\n   %s" %\
            (",".join(options.transformers), params["options"])

    outfile.write(RST_TEMPLATE % params)


def writeNotebook(outfile, options, kwargs,
                  renderer_options,
                  transformer_options,
                  display_options,
                  modulename, name):
    '''write a snippet to paste with the ipython notebook.
    '''

    cmd_options = [
        'do_print = False',
        'tracker="%s"' % options.tracker,
        'renderer="%s"' % options.renderer,
        'trackerdir="%s"' % options.trackerdir,
        'workdir="%s"' % os.getcwd()]

    for key, val in list(kwargs.items()) +\
            list(renderer_options.items()) +\
            list(transformer_options.items()):
        if val is None:
            cmd_options.append("%s" % key)
        else:
            if Utils.isString(val):
                cmd_options.append('%s="%s"' % (key, val))
            else:
                cmd_options.append('%s=%s' % (key, val))
    if options.transformers:
        cmd_options.append(
            "transformer=['%s']" % "','".join(options.transformers))

    # no module name in tracker
    params = {"tracker": "%s" % (name),
              "options": ",\n".join(cmd_options)}

    outfile.write(Utils.NOTEBOOK_TEXT_TEMPLATE % params)


def run(name, t, kwargs):

    print("%s: collecting data started" % name)
    t(**kwargs)
    print("%s: collecting data finished" % name)


def main(argv=None, **kwargs):
    '''main function for test.py.

    Long-form of command line arguments can also be supplied as kwargs.

    If argv is not None, command line parsing will be performed.
    '''
    parser = optparse.OptionParser(version="%prog version: $Id$",
                                   usage=globals()["__doc__"])

    parser.add_option("-t", "--tracker", dest="tracker", type="string",
                      help="tracker to use [default=%default]")

    parser.add_option("-p", "--page", dest="page", type="string",
                      help="render an rst page [default=%default]")

    parser.add_option("-a", "--tracks", dest="tracks", type="string",
                      help="tracks to use [default=%default]")

    parser.add_option("-m", "--transformer", dest="transformers",
                      type="string", action="append",
                      help="add transformation [default=%default]")

    parser.add_option("-s", "--slices", dest="slices", type="string",
                      help="slices to use [default=%default]")

    parser.add_option("-r", "--renderer", dest="renderer", type="string",
                      help="renderer to use [default=%default]")

    parser.add_option("-w", "--path", "--trackerdir",
                      dest="trackerdir", type="string",
                      help="path to trackers [default=%default]")

    parser.add_option("-f", "--force", dest="force", action="store_true",
                      help="force recomputation of data by deleting cached "
                      "results [default=%default]")

    parser.add_option("-o", "--option", dest="options", type="string",
                      action="append",
                      help="renderer options - supply as key=value pairs "
                      "(without spaces). [default=%default]")

    parser.add_option("-l", "--language", dest="language", type="choice",
                      choices=("rst", "notebook"),
                      help="output language for snippet. Use ``rst`` "
                      "to create a snippet to paste "
                      "into a cgatreport document. Use ``notebook`` to "
                      "create a snippet to paste "
                      "into an ipython notebook [default=%default]")

    parser.add_option("--no-print", dest="do_print", action="store_false",
                      help="do not print an rst text element to create "
                      "the displayed plots [default=%default].")

    parser.add_option("--no-show", dest="do_show", action="store_false",
                      help="do not show a plot [default=%default].")

    parser.add_option("--layout", dest="layout", type="string",
                      help="output rst with layout [default=%default].")

    parser.add_option("-i", "--start-interpreter", dest="start_interpreter",
                      action="store_true",
                      help="do not render, but start python interpreter "
                      "[default=%default].")

    parser.add_option("-I", "--ii", "--start-ipython", dest="start_ipython",
                      action="store_true",
                      help="do not render, start ipython interpreter "
                      "[default=%default].")

    parser.add_option(
        "--workdir", dest="workdir", type="string",
        help="working directory - change to this directory "
        "before executing "
        "[default=%default]")

    parser.add_option(
        "--hardcopy", dest="hardcopy", type="string",
        help="output images of plots. The parameter should "
        "contain one or more %s "
        "The suffix determines the type of plot. "
        "[default=%default].")

    parser.set_defaults(
        loglevel=1,
        tracker=None,
        transformers=[],
        tracks=None,
        slices=None,
        options=[],
        renderer="table",
        do_show=True,
        do_print=True,
        force=False,
        trackerdir=TRACKERDIR,
        caption="add caption here",
        start_interpreter=False,
        start_ipython=False,
        language="rst",
        workdir=None,
        layout=None,
        dpi=100)

    if argv is None and len(kwargs) == 0:
        argv = sys.argv

    if argv:
        (options, args) = parser.parse_args(argv)

        if len(args) == 2:
            options.tracker, options.renderer = args

    else:
        (options, args) = parser.parse_args([])

        ######################################################
        # set keyword arguments as options
        for keyword, value in kwargs.items():
            if hasattr(options, keyword):
                setattr(options, keyword, value)
                del kwargs[keyword]

        # change some kwarguments
        if options.transformers:
            for keyword, value in kwargs.items():
                if keyword.startswith("tf"):
                    kwargs["tf-{}".format(keyword[2:])] = value

    if options.workdir is not None:
        savedir = os.getcwd()
        os.chdir(options.workdir)
        Utils.getParameters(sorted(glob.glob("*.ini")))
    else:
        savedir = None

    ######################################################
    # configure options
    options.trackerdir = os.path.abspath(
        os.path.expanduser(options.trackerdir))
    if os.path.exists(options.trackerdir):
        sys.path.insert(0, options.trackerdir)
    else:
        L.warn("directory %s does not exist" % options.trackerdir)

    ######################################################
    # test plugins
    for x in options.options:
        if "=" in x:
            data = x.split("=")
            key, val = [y.strip() for y in (data[0], "=".join(data[1:]))]
        else:
            key, val = x.strip(), None
        kwargs[key] = val

    if options.tracks:
        kwargs["tracks"] = options.tracks
    if options.slices:
        kwargs["slices"] = options.slices

    kwargs = Utils.updateOptions(kwargs)

    option_map = getOptionMap()
    renderer_options = Utils.selectAndDeleteOptions(
        kwargs, option_map["render"])
    transformer_options = Utils.selectAndDeleteOptions(
        kwargs, option_map["transform"])
    display_options = Utils.selectAndDeleteOptions(
        kwargs, option_map["display"])

    ######################################################
    # decide whether to render or not
    if options.renderer == "none" or options.start_interpreter or \
       options.start_ipython or options.language == "notebook":
        renderer = None
    else:
        renderer = Utils.getRenderer(options.renderer, renderer_options)

    transformers = Utils.getTransformers(
        options.transformers, transformer_options)

    exclude = set(("Tracker",
                   "TrackerSQL",
                   "returnLabeledData",
                   "returnMultipleColumnData",
                   "returnMultipleColumns",
                   "returnSingleColumn",
                   "returnSingleColumnData",
                   "SQLError",
                   "MultipleColumns",
                   "MultipleColumnData",
                   "LabeledData",
                   "DataSimple",
                   "Data"))

    ######################################################
    # build from tracker
    if options.tracker:

        if "." in options.tracker:
            parts = options.tracker.split(".")
            tracker_modulename = ".".join(parts[:-1])
            tracker_name = parts[-1]
        else:
            tracker_modulename = None
            tracker_name = options.tracker

        try:
            _code, tracker, tracker_path = Utils.makeTracker(
                options.tracker, (), kwargs)
        except ImportError:
            # try to find class in module
            trackers = []

            for filename in glob.glob(
                    os.path.join(options.trackerdir, "*.py")):
                modulename = os.path.basename(filename)
                trackers.extend(
                    [x for x in getTrackers(modulename)
                     if x[0] not in exclude])

            for name, tracker_class, modulename, is_derived in trackers:
                if name == tracker_name:
                    if tracker_modulename is not None:
                        if modulename == tracker_modulename:
                            break
                    else:
                        tracker_modulename = modulename
                        break
            else:
                available_trackers = set([x[0] for x in trackers if x[3]])
                print(
                    "unknown tracker '%s': possible trackers are\n  %s" %
                    (options.tracker, "\n  ".join(sorted(available_trackers))))
                print(
                    "(the list above does not contain functions).")
                sys.exit(1)

            # instantiate functors
            if is_derived:
                tracker = tracker_class(**kwargs)
            #  but not functions
            else:
                tracker = tracker_class

        # remove everything related to that tracker for a clean slate
        if options.force:
            removed = CGATReport.clean.removeTracker(tracker_name)
            print("removed all data for tracker %s: %i files" %
                  (tracker_name, len(removed)))

        dispatcher = Dispatcher(tracker, renderer, transformers)

        if renderer is None:
            # dispatcher.parseArguments(**kwargs)
            # result = dispatcher.collect()
            # result = dispatcher.transform()
            result = dispatcher(**kwargs)
            options.do_print = options.language == "notebook"
            options.do_show = False
            options.hardcopy = False
        else:
            # needs to be resolved between renderer and dispatcher options
            result = dispatcher(**kwargs)

        if options.do_print:

            sys.stdout.write(".. ---- TEMPLATE START --------\n\n")

            if options.language == "rst":
                writeRST(sys.stdout,
                         options,
                         kwargs,
                         renderer_options,
                         transformer_options,
                         display_options,
                         tracker_modulename,
                         tracker_name)
            elif options.language == "notebook":
                writeNotebook(sys.stdout,
                              options,
                              kwargs,
                              renderer_options,
                              transformer_options,
                              display_options,
                              tracker_modulename,
                              tracker_name)

            sys.stdout.write("\n.. ---- TEMPLATE END ----------\n")

        sys.stdout.write("\n.. ---- OUTPUT-----------------\n")

        if result and renderer is not None:
            if options.layout is not None:
                lines = Utils.layoutBlocks(result, layout=options.layout)
                print "\n".join(lines)
            else:
                for r in result:
                    if r.title:
                        print ("")
                        print ("title: %s" % r.title)
                        print ("")
                    for s in r:
                        print(str(s))

        if options.hardcopy:

            fig_managers = _pylab_helpers.Gcf.get_all_fig_managers()
            # create all the images
            for figman in fig_managers:
                # create all images
                figid = figman.num
                outfile = re.sub("%s", str(figid), options.hardcopy)
                figman.canvas.figure.savefig(outfile, dpi=options.dpi)

        if result and options.do_show:
            if options.renderer.startswith("r-"):
                for rr in result:
                    for r in rr:
                        if hasattr(r, 'rggplot'):
                            from rpy2.robjects import r as R
                            import rpy2.rinterface
                            try:
                                R.plot(r.rggplot)
                            except rpy2.rinterface.RRuntimeError, msg:
                                if re.search("object.*not found", str(msg)):
                                    print '%s: available columns in dataframe=%s' % \
                                        (msg,
                                          R('''colnames(rframe)'''))

                print("press Ctrl-c to stop")
                while 1:
                    pass

            elif len(_pylab_helpers.Gcf.get_all_fig_managers()) > 0:
                plt.show()

            else:
                for rr in result:
                    for r in rr:
                        if hasattr(r, 'xls'):
                            tmpfile, outpath = tempfile.mkstemp(
                                dir='.', suffix='.xlsx')
                            os.close(tmpfile)
                            print ('saving xlsx to %s' % outpath)
                            r.xls.save(outpath)
                        elif hasattr(r, 'bokeh'):
                            import bokeh.plotting as bk
                            bk.show()

    ######################################################
    # build page
    elif options.page:

        from CGATReport import build
        CGATReport.report_directive.DEBUG = True
        CGATReport.report_directive.FORCE = True

        if not os.path.exists(options.page):
            raise IOError("page %s does not exist" % options.page)

        options.num_jobs = 1

        build.buildPlots(
            [options.page, ], options, [], os.path.dirname(options.page))

        if options.do_show:
            if options.renderer.startswith("r-"):
                print("press Ctrl-c to stop")
                while 1:
                    pass

            elif _pylab_helpers.Gcf.get_all_fig_managers() > 0:
                plt.show()

    else:
        raise ValueError(
            "please specify either a tracker "
            "(-t/--tracker) or a page (-p/--page) to test")

    if savedir is not None:
        os.chdir(savedir)

    if options.tracker and renderer is None:
        datatree = dispatcher.getDataTree()
        dataframe = dispatcher.getDataFrame()

        # trying to push R objects
        # from rpy2.robjects import r as R
        # for k, v in flat_iterator(datatree):
        #     try:
        #         R.assign(k, v)
        #     except ValueError, msg:
        #         print ("could not push %s: %s" % (k,msg))
        #         pass
        # print ("----------------------------------------")
        if options.start_interpreter:
            print ("--> cgatreport - available data structures <--")
            print ("    datatree=%s" % type(datatree))
            print ("    dataframe=%s" % type(dataframe))
            interpreter = code.InteractiveConsole(
                dict(globals().items() + locals().items()))
            interpreter.interact()
            return dataframe
        elif options.start_ipython:
            import IPython
            IPython.embed()
            return dataframe

        return dataframe

if __name__ == "__main__":
    sys.exit(main(argv=sys.argv))
